/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <iostream>
#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Source.h>
#include <Map.h>
#include <TimeStamp.h>
#include <Stream.h>
#include <math.h>

#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <iomanip>
#include <sstream>

//#ifdef ENABLE_WEBSOCKET

#include "WebSocket.hpp"
#include <boost/shared_ptr.hpp>

//#include "websocketpp/roles/server.hpp"
#include "websocketpp/roles/client.hpp"
#include "websocketpp/websocketpp.hpp"

#include <boost/algorithm/string/replace.hpp>
#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <boost/thread.hpp>
//#include "SubscriberWebSocket.hpp"

#include <map>
#include <string>
#include <queue>

//#include "HarkParamsObjects.h"

using namespace FD;
using namespace std;

using boost::asio::ip::tcp;
using websocketpp::server;
using websocketpp::client;

//using namespace websocketconnector;
//using namespace websocketconnector::ws_client_handler;
//using namespace websocketconnector::ws_server_handler;

class SubscriberWebSocket;

DECLARE_NODE(SubscriberWebSocket)
/*Node
 *
 * @name SubscriberWebSocket
 * @category HARKD:AudioIO
 * @description Many stream data (Ex: Multi-channel sound, FFT, Source ID, etc...) from the WebSocket network.
 *
 * @parameter_name ENABLE_STATUS_DEBUG
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description status print flac decoder data of this module in case of true.
 *
 * @parameter_name LENGTH
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description The frame length of each channel (in samples) [default: 512].
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The shift length beween adjacent frames (in samples)[default: 160].
 *
 * @parameter_name CHANNEL
 * @parameter_type int
 * @parameter_value 8
 * @parameter_description The number of microphone.
 *
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description Sampling rate (Hz) [default: 16000].
 *
 * @parameter_name NUM_OF_STREAM
 * @parameter_type int
 * @parameter_value 1
 * @parameter_list 1:2:3:4:5:6:7:8:9:10:11:12:13:14:15:16
 * @parameter_description Number of in/out stream data.
 *
 * @parameter_name STREAM_SELECTION_0
 * @parameter_type string
 * @parameter_value DataStream
 * @parameter_list ParamStream:DataStream
 * @parameter_valid NUM_OF_STREAM=1:2:3:4:5:6:7:8:9:10:11:12:13:14:15:16
 * @parameter_description Selector of stream you want to realize. All other subscription will be disabled.
 *
 * @parameter_name STREAM0_OVERLAPCONTROL
 * @parameter_type string
 * @parameter_value Through
 * @parameter_list Through:Overlap:DisOverlap:Auto
 * @parameter_valid STREAM_SELECTION_0=DataStream
 * @parameter_description If you want to force overlap, please select to Overlap. If you want to no processing, please select to Through. You can't select Auto, because this function is pending.
 *
 * @output_name OUT_0
 * @output_type any
 * @output_description Output of STREAM_SELECTION_0.
 *
 * @input_name IN_CONDITION_SYNC
 * @input_type bool
 * @input_description If the socket connection failed or the synchronize other node was changed to false condition will false. And if this input was not connected ignore other node condition.
 *
 * @output_name OUT_CONDITION
 * @output_type bool
 * @output_description If the socket connection failed condition will false. Normally always true.
 *

 END*/

class SubscriberWebSocket : public BufferedNode {
  //  Output IDs | Connected Flags
  int outputID[16];  bool boutputID[16];

  // Condition in/out
  int input_condition_syncID; bool binput_condition_syncID;
  int output_conditionID;

  // Parameters [b:bool, i:int, f:float, d:double, s:string, o:ObjectRef, cx:Complex<float>, vx:Vector<x>, mx:Matrix<x>, px:Map<int, x>]
  bool   param_b_debug_status;  // Enable debug status         [true:enable, false:disable]
  bool   param_b_debug_sync;    // Enable debug synchronize    [true:enable, false:disable]
  bool   param_b_mode_server;   // Select Server/Client mode   [true:server, false:client]
  bool   param_b_mode_ssl;      // Use SSL connection          [true:enable, false:disable]
  bool   param_b_mode_dosguard; // Use DoS attack guard        [true:enable, false:disable]
  int    param_i_limit_size;    // Receive data limit size for DoS attack guard function
  string param_s_host;          // IP or Hostname
  int    param_i_port;          // Port number
  string param_s_dir;           // Directory suffix
  bool   param_b_timeout;       // Use timeout for client mode [true:enable, false:disable]
  bool   param_i_timeout;       // Timeout[ms] for timeout function
  int    param_i_buffer;        // Buffer size

  int    param_i_length;        // Frame length
  int    param_i_advance;       // Shift size
  int    param_i_channels;      // Number of channels
  int    param_i_sampling_rate; // Sampling rate

  int    param_i_streams;       // Number of target streams
  int    param_i_stream_type[16]; // Stream type for each stream

  bool   param_b_overlap[16]; bool boutput_overlap[16]; bool boutput_disoverlap[16]; bool boutput_autooverlap[16];
/*/*
  /*int c_sock;         // եǥץ
  int sock;
  int yes;
  struct sockaddr_in addr;
  struct sockaddr_in client;
  short int *buffer;  // Хåե
  int buf_size;       // Хåե
  
  int isConnected;
*/

  int header_magic;      // Magic Code, If you read this magic code by int on 32bit system
  int header_padding;    // for 64bit system guard bits, Maybe 64bit system is if you are reading along with the padding when you read the magic code.
  int host_sz_char;
  int host_sz_int;
  int host_sz_float;
  int host_sz_double;
  bool b_host_little_endian;
  string header_string;

  int overlap_size;

  bool socket_condition;

  int dummy;

  // for client
  //boost::shared_ptr<websocketconnector::ws_client_handler_ptr> m_cli_handler;
  websocketconnector::ws_client_handler_ptr m_cli_handler;
  websocketpp::client *m_cli_endpoint;
  websocketpp::client::connection_ptr m_connect_ptr;

  // for server
  websocketconnector::ws_server_handler_ptr m_srv_handler;
  websocketpp::server *m_srv_endpoint;

  // for multi thread function ... pending
  boost::shared_ptr<boost::thread> m_thread_hark;

  public:
  // Constructor
  DecodeFLAC(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params),
    input_condition_syncID(-1), binput_condition_syncID(false),
    socket_condition(true),
    m_cli_endpoint(NULL),m_srv_endpoint(NULL),
    dummy(0)
  {
    for(int stream=0; stream<16; stream++){
      outputID[stream]            = -1;
      boutputID[stream]           = false;
      boutput_overlap[stream]     = false;
      boutput_disoverlap[stream]  = false;
      boutput_autooverlap[stream] = false;
    }
    host_sz_char   = sizeof(char);   // Size of char
    host_sz_int    = sizeof(int);    // Size of int
    host_sz_float  = sizeof(float);  // Size of float   ** Size of complex float considers (float * 2)
    host_sz_double = sizeof(double); // Size of double  ** Size of complex double considers (double * 2)

    output_conditionID = addOutput("OUT_CONDITION");

    param_b_debug_status = dereference_cast<bool>(parameters.get("ENABLE_STATUS_DEBUG"));
    param_b_debug_sync = dereference_cast<bool>(parameters.get("ENABLE_SYNC_DEBUG"));
    param_b_mode_server = dereference_cast<bool>(parameters.get("SERVER_MODE"));
    param_b_mode_ssl = dereference_cast<bool>(parameters.get("USE_SSL"));
    param_b_mode_dosguard = dereference_cast<bool>(parameters.get("USE_DOS_ATTACK_GUARD"));
    // Only enable when the USE_DOS_ATTACK_GUARD is ture.
    if(parameters.exist("LIMIT")){ param_i_limit_size = dereference_cast<int>(parameters.get("LIMIT")); }
    else{                          param_i_limit_size = 0;                                              }
    param_s_host = object_cast<String>(parameters.get("PEER_HOST"));
    param_i_port = dereference_cast<int>(parameters.get("PEER_PORT"));
    param_s_dir = object_cast<String>(parameters.get("DIR_SUFFIX"));
    // Only enable when the SERVER_MODE is true(client).
    if(parameters.exist("USE_TIMEOUT")){ param_b_timeout = dereference_cast<bool>(parameters.get("USE_TIMEOUT")); }
    else{                                param_b_timeout = false;                                                 }
    // Only enable when the USE_TIMEOUT is ture.
    if(parameters.exist("TIMEOUT")){ param_i_timeout = dereference_cast<int>(parameters.get("TIMEOUT")); }
    else{                            param_i_timeout = 0;                                                }
    param_i_buffer = dereference_cast<int>(parameters.get("BUFFER_SIZE"));
    param_i_length = dereference_cast<int>(parameters.get("LENGTH"));
    param_i_advance = dereference_cast<int>(parameters.get("ADVANCE"));
    if(param_i_advance < param_i_length){      overlap_size = param_i_length - param_i_advance; }
    else if(param_i_advance > param_i_length){ overlap_size = 0; param_i_advance = param_i_length; cout << "ADVANCE reset value from LENGTH." << endl; }
    else{                                      overlap_size = 0; }
    if(overlap_size == 0){ // Reset overlap flags, because "LENGTH <= ADVANCE" or disabled overlap.
      for(int stream=0; stream<16; stream++){
        param_b_overlap[stream] = false;
        boutput_overlap[stream] = false;
        boutput_disoverlap[stream] = false;
        boutput_autooverlap[stream] = false;
      }
    }
    else{
    // param_b_overlap?     : If the this flag is true, indicates already read from that the parameter request. Show that it is initialized with error if false.
    // boutput?_overlap     : If the this flag is true, indicates that there is a request for overlap. May change during the execution of each frame.
    // boutput?_disoverlap  : If the this flag is true, indicates that there is a request for dis-overlap. May change during the execution of each frame.
    // boutput?_autooverlap : If the this flag is true, the overlap flag is decided at runtime.

      string param_s_temp  = "";
      string param_s_temp0 = "";
      string param_s_temp1 = "";
      for(int stream=0; stream<16; stream++){
        switch(stream){
        case 0:  param_s_temp0 = "STREAM_SELECTION_0";  param_s_temp1 = "STREAM0_OVERLAPCONTROL";  break;
        case 1:  param_s_temp0 = "STREAM_SELECTION_1";  param_s_temp1 = "STREAM1_OVERLAPCONTROL";  break;
        case 2:  param_s_temp0 = "STREAM_SELECTION_2";  param_s_temp1 = "STREAM2_OVERLAPCONTROL";  break;
        case 3:  param_s_temp0 = "STREAM_SELECTION_3";  param_s_temp1 = "STREAM3_OVERLAPCONTROL";  break;
        case 4:  param_s_temp0 = "STREAM_SELECTION_4";  param_s_temp1 = "STREAM4_OVERLAPCONTROL";  break;
        case 5:  param_s_temp0 = "STREAM_SELECTION_5";  param_s_temp1 = "STREAM5_OVERLAPCONTROL";  break;
        case 6:  param_s_temp0 = "STREAM_SELECTION_6";  param_s_temp1 = "STREAM6_OVERLAPCONTROL";  break;
        case 7:  param_s_temp0 = "STREAM_SELECTION_7";  param_s_temp1 = "STREAM7_OVERLAPCONTROL";  break;
        case 8:  param_s_temp0 = "STREAM_SELECTION_8";  param_s_temp1 = "STREAM8_OVERLAPCONTROL";  break;
        case 9:  param_s_temp0 = "STREAM_SELECTION_9";  param_s_temp1 = "STREAM9_OVERLAPCONTROL";  break;
        case 10: param_s_temp0 = "STREAM_SELECTION_10"; param_s_temp1 = "STREAM10_OVERLAPCONTROL"; break;
        case 11: param_s_temp0 = "STREAM_SELECTION_11"; param_s_temp1 = "STREAM11_OVERLAPCONTROL"; break;
        case 12: param_s_temp0 = "STREAM_SELECTION_12"; param_s_temp1 = "STREAM12_OVERLAPCONTROL"; break;
        case 13: param_s_temp0 = "STREAM_SELECTION_13"; param_s_temp1 = "STREAM13_OVERLAPCONTROL"; break;
        case 14: param_s_temp0 = "STREAM_SELECTION_14"; param_s_temp1 = "STREAM14_OVERLAPCONTROL"; break;
        case 15: param_s_temp0 = "STREAM_SELECTION_15"; param_s_temp1 = "STREAM15_OVERLAPCONTROL"; break;
        default: param_s_temp0 = "STREAM_SELECTION_15"; param_s_temp1 = "STREAM15_OVERLAPCONTROL"; break;
        }
        if(parameters.exist(param_s_temp0)){
          param_s_temp = object_cast<String>(parameters.get(param_s_temp0));
          if(param_s_temp == "DataStream"){
            param_i_stream_type[stream] = 0; // input_type = 0x10000000;
            if(parameters.exist(param_s_temp1)){
              param_s_temp = object_cast<String>(parameters.get(param_s_temp1));
              if(param_s_temp == "Through"){         param_b_overlap[stream] = true;  boutput_overlap[stream] = false; boutput_disoverlap[stream] = false; boutput_autooverlap[stream] = false; }
              else if(param_s_temp == "Overlap"){    param_b_overlap[stream] = true;  boutput_overlap[stream] = true;  boutput_disoverlap[stream] = false; boutput_autooverlap[stream] = false; }
              else if(param_s_temp == "DisOverlap"){ param_b_overlap[stream] = true;  boutput_overlap[stream] = false; boutput_disoverlap[stream] = true;  boutput_autooverlap[stream] = false; }
              else if(param_s_temp == "Auto"){       param_b_overlap[stream] = true;  boutput_overlap[stream] = false; boutput_disoverlap[stream] = false; boutput_autooverlap[stream] = true;  }
              else{                                  param_b_overlap[stream] = false; boutput_overlap[stream] = false; boutput_disoverlap[stream] = false; boutput_autooverlap[stream] = false; cerr << "Unknown overlap control type. [" << param_s_temp << "]" << endl; }
            }
            else{
              cerr << "Parameter is not exist. [STREAM1_OVERLAPCONTROL]" << endl;
              param_b_overlap[stream] = false; boutput_overlap[stream] = false; boutput_disoverlap[stream] = false; boutput_autooverlap[stream] = false;
            }
          }
          else if(param_s_temp == "ParamStream"){
            param_i_stream_type[stream] = 1; // input_type = 0x20000000; // b_stream1_isdata = false;
          }
        }
      }
    }

    param_i_channels = dereference_cast<int>(parameters.get("CHANNEL"));
    param_i_sampling_rate = dereference_cast<int>(parameters.get("SAMPLING_RATE"));
    param_i_streams = dereference_cast<int>(parameters.get("NUM_OF_STREAM"));

    if(param_i_streams <   1){ cerr << "Parameter error : NUM_OF_STREAM(Out of range) " << param_i_streams << endl; param_i_streams = 1; }
    if(param_i_streams >  16){ cerr << "Parameter error : NUM_OF_STREAM(Out of range) " << param_i_streams << endl; param_i_streams = 16; }
    for(int stream=param_i_streams-1; stream>=0; stream--){ /*param_s_stream1  = object_cast<String>(parameters.get("STREAM_SELECTION_1")); */ }


    // Start connection (WebSocket open)
    if(param_b_mode_server){
      try{
        //// create an instance of handler ////
        //websocketconnector::server_handler::startServer(); // temp function define. normally ****->startServer();
        //server::handler::ptr handler(new ws_server_handler()); 
        //ws_server_handler_ptr handler(new ws_server_handler());
        m_srv_handler.reset(new websocketconnector::ws_server_handler());

        //// create a server the listen on port uses handler ////
        //websocketpp::server endpoint(handler);
        m_srv_endpoint = (new websocketpp::server(m_srv_handler));

        // set log level for message print
        //srv_endpoint.alog().set_level(websocketpp::log::alevel::CONNECT);
        //srv_endpoint.alog().set_level(websocketpp::log::alevel::DISCONNECT);
        //srv_endpoint.elog().set_level(websocketpp::log::elevel::RERROR);
        //srv_endpoint.elog().set_level(websocketpp::log::elevel::FATAL);
        m_srv_endpoint->alog().set_level(websocketpp::log::alevel::CONNECT);
        m_srv_endpoint->alog().set_level(websocketpp::log::alevel::DISCONNECT);
        m_srv_endpoint->elog().set_level(websocketpp::log::elevel::RERROR);
        m_srv_endpoint->elog().set_level(websocketpp::log::elevel::FATAL);

        if(param_i_limit_size > 0){
          //server->set_max_messages_size(param_i_limit_size);
          std::cout << "Set receive message limitation [" << param_i_limit_size << " Byte]." << std::endl;
        }

        //m_srv_endpoint->listen(param_i_port);
        //std::cout << "Starting WebSocket server on port [" << param_i_port << "]." << std::endl;

        //// start thread ////
        //boost::thread thread_hark(boost::bind(&client::run, &cli_endpoint, false));
        //m_thread_hark.reset(new boost::thread(boost::bind(&websocketpp::role::server<websocketpp::server>::listen, m_srv_endpoint, param_i_port, 1)));
        void(websocketpp::role::server<websocketpp::server>::*f)(uint16_t, size_t) = &websocketpp::role::server<websocketpp::server>::listen;
        m_thread_hark.reset(new boost::thread(boost::bind(f, m_srv_endpoint, param_i_port, 1)));
        std::cout << "Starting WebSocket server on port [" << param_i_port << "]." << std::endl;
      }
      catch(exception& e){
        std::cerr << "Exception: " << e.what() << std::endl;
      }
    }
    else{
      std::stringstream ss_uri;
      ss_uri << "ws://" << param_s_host << ":" << param_i_port << "/" << param_s_dir;
      std::string uri = ss_uri.str();
      //std::string uri = "ws://localhost:8003/hark";
      try{
        //// create an instance of handler ////
        //websocketconnector::client_handler::startClient(); // temp function define. normally ****->startClient();
        //websocketconnector::ws_client_handler_ptr cli_handler(new websocketconnector::ws_client_handler());
        //client::handler::ptr handler(new websocketconnector::ws_client_handler());
        m_cli_handler.reset(new websocketconnector::ws_client_handler());

        //// create a client uses handler ////
        //websocketpp::client cli_endpoint(cli_handler);
        m_cli_endpoint = (new websocketpp::client(m_cli_handler));

        //websocketpp::client::connection_ptr connect_ptr;

        //// set log level for message print ////
        //cli_endpoint.alog().unset_level(websocketpp::log::alevel::ALL); // disable all infomation message
        //cli_endpoint.elog().unset_level(websocketpp::log::elevel::ALL); // disable all error message
        //cli_endpoint.elog().set_level(websocketpp::log::elevel::RERROR);
        //cli_endpoint.elog().set_level(websocketpp::log::elevel::FATAL);
        m_cli_endpoint->alog().unset_level(websocketpp::log::alevel::ALL); // disable all infomation message
        m_cli_endpoint->elog().unset_level(websocketpp::log::elevel::ALL); // disable all error message
        m_cli_endpoint->elog().set_level(websocketpp::log::elevel::RERROR);
        m_cli_endpoint->elog().set_level(websocketpp::log::elevel::FATAL);

        //// get connection ////
        //connect_ptr = cli_endpoint.get_connection(uri);
        //connect_ptr->add_request_header("User-Agent", "WebSocket++/0.2.0 FlowDesignerPublisherWebSocket/0.0.1");
        //connect_ptr->add_subprotocol("hri.websocketpp.hark");
        //connect_ptr->set_origin("http://hark.jp.honda-ri.com");
        //cli_endpoint.connect(connect_ptr);
        m_connect_ptr = m_cli_endpoint->get_connection(uri);
        m_connect_ptr->add_request_header("User-Agent", "WebSocket++/0.2.0 FlowDesignerPublisherWebSocket/0.0.1");
        m_connect_ptr->set_origin("http://hark.jp.honda-ri.com");
        m_cli_endpoint->connect(m_connect_ptr);

        //// start thread ////
        //boost::thread thread_hark(boost::bind(&client::run, &cli_endpoint, false));
        m_thread_hark.reset(new boost::thread(boost::bind(&client::run, m_cli_endpoint, false)));
        std::cout << "Starting WebSocket client." << std::endl;

        //// send test code (obsolete) ////
        //char test_buffer[2048];
        //while(cin.getline(test_buffer, 2048)){ cli_handler->send(test_buffer); }
        //// waiting thread finish for send test code (obsolete) ////
        //thread_hark.join();
      }
      catch(std::exception& e){
        std::cerr << "Exception: " << e.what() << std::endl;
      }
    }



/*


    length = dereference_cast<int>(parameters.get("LENGTH"));
    if (parameters.exist("ADVANCE")) {
      advance = dereference_cast<int>(parameters.get("ADVANCE"));
    }
    else advance = length;
        
    num_channels = dereference_cast<int>(parameters.get("CHANNEL"));
    port = dereference_cast<int>(parameters.get("PORT"));
    host = object_cast<String>(parameters.get("PEER_HOST"));

    buf_size = 16000 * 16 * 10;        
    buffer = (short int*)calloc(sizeof(short int), buf_size);

	  c_sock = socket(AF_INET, SOCK_STREAM, 0);
	  if (c_sock < 0) {
		perror("building socket failed");
		exit(-1);
	  }
	  addr.sin_family = AF_INET;
	  addr.sin_port = htons(port);
	  addr.sin_addr.s_addr = INADDR_ANY;
	  
	  setsockopt(c_sock, SOL_SOCKET, SO_REUSEADDR,
				 (const char*)&yes, sizeof(yes));
	yes = 1;
*/
    inOrder = true;
    std::cout << getName() << " Constructed..." << std::endl;
  }
    
  ~DecodeFLAC()
  {
    // Stop connection (WebSocket close)
    if(param_b_mode_server){
      //websocketconnector::ws_server_handler::stopServer(); // temp function define. normally ****->stopServer();
      if(m_thread_hark){ m_thread_hark->join(); }
    }
    else{
      //websocketconnector::ws_client_handler::stopClient(); // temp function define. normally ****->stopClient();
      if(m_thread_hark){ m_thread_hark->join(); }
    }
    std::cout << getName() << " Destructed..." << std::endl;
  }

  // Initialize
  virtual void initialize() {

    for(int stream=15; stream>=0; stream--){
      if(boutputID[stream]){
        outputs[output_conditionID].lookAhead = max(outputs[outputID[stream]].lookAhead, outputs[output_conditionID].lookAhead);
        outputs[output_conditionID].lookBack  = max(outputs[outputID[stream]].lookBack,  outputs[output_conditionID].lookBack);
      }
    }
    outputs[output_conditionID].lookAhead = 1+outputs[output_conditionID].lookAhead;
    outputs[output_conditionID].lookBack  = 1+outputs[output_conditionID].lookBack;
    for(int stream=15; stream>=0; stream--){
      if(boutputID[stream]){
        outputs[outputID[stream]].lookAhead = outputs[output_conditionID].lookAhead;
      }
    }

    // I know that it is not desirable that follows "this->BufferedNode::initialize();".
    // However, "this->BufferedNode::initialize();" performed as a work-around for slow processing.
    int countout = 0;
    if(param_b_mode_server){
      // Server listen waiting loop recommend infinite loop. If you need terminative loop by timeout, change to use this lines.
      //std::cout << "waiting client connection...";
      //while(!(m_srv_handler->get_connect_state())){ std::cout << "."; countout++; if(countout > 0x00FFFFFF){ throw new NodeException(this, "connection timeout.", __FILE__, __LINE__); } }
      //std::cout << std::endl;
      std::cout << "waiting client connection...";
      while(!(m_srv_handler->get_connect_state())){ std::cout << "."; }
      std::cout << std::endl;
    }
    else{
      std::cout << "waiting server connection...";
      while(!(m_cli_handler->get_connect_state())){ std::cout << "."; countout++; if(countout > 0x00FFFFFF){ throw new NodeException(this, "connection timeout.", __FILE__, __LINE__); } }
      std::cout << std::endl;
    }

    this->BufferedNode::initialize();
    std::cout << getName() << " Initialized..." << std::endl;
  }
    


  // dynamic input-port translation
  virtual int translateInput(string inputName) {
    // check whether the input ports have arc from other module.
    for(unsigned int i = 0; i < inputs.size(); i++) {
      if(inputs[i].name == inputName) {
        return i;
      }
    }
    if(inputName == "IN_CONDITION_SYNC"){ binput_condition_syncID = true; return input_condition_syncID = addInput(inputName); }
    else{ throw new NodeException(this, inputName + " is not supported.", __FILE__, __LINE__); }
  }

  // dynamic output-port translation
  virtual int translateOutput (string outputName)
  {
    if(outputName == "OUT_PCM"){
      if(boutputID[0]){          return outputID[0]; }
      else{ boutputID[0] = true; return outputID[0] = addOutput(outputName); }
    }
    else if(outputName == "OUT_1"){
      if(boutputID[1]){          return outputID[1]; }
      else{ boutputID[1] = true; return outputID[1] = addOutput(outputName); }
    }
    else if(outputName == "OUT_2"){
      if(boutputID[2]){          return outputID[2]; }
      else{ boutputID[2] = true; return outputID[2] = addOutput(outputName); }
    }
    else if(outputName == "SOURCE"){
      if(boutputID[3]){          return outputID[3]; }
      else{ boutputID[3] = true; return outputID[3] = addOutput(outputName); }
    }

    for(unsigned int i=0; i<outputNames.size(); i++){
      if(outputNames[i] == outputName){ return i; }
    }  

    return addOutput(outputName);
  }

  void calculate(int output_id, int count, Buffer &out){
//    // update websocket condition flag
//    if(param_b_mode_server){
//      if(m_thread_hark){
//        if(m_srv_handler->get_connect_state()){ socket_condition = true;  }
//        else{                                   socket_condition = false; }
//      }
//      else{
//        socket_condition = false;
//      }
//    }
//    else{
//      if(m_thread_hark){
//        if(m_cli_handler->get_connect_state()){ socket_condition = true;  }
//        else{                                   socket_condition = false; }
//      }
//      else{
//        socket_condition = false;
//      }
//    }
//    // control condition flag
//    Buffer &output_conditionBuffer = *(outputs[output_conditionID].buffer);
//    if(binput_condition_syncID){
//      ObjectRef input_condition_obj = getInput(input_condition_syncID, count);
//      bool input_condition = dereference_cast<bool> (input_condition_obj);
//      if(input_condition && socket_condition){ output_conditionBuffer[count] = TrueObject;  }
//      else{                                    output_conditionBuffer[count] = FalseObject; }
//    }
//    else{
//      if(socket_condition){ output_conditionBuffer[count] = TrueObject;  }
//      else{                 output_conditionBuffer[count] = FalseObject; }
//    }
//
//    bool nullpo=false;
//    std::string buffer(""); buffer.reserve(4000000);
//    if(param_b_mode_server){
//      if(m_srv_handler != NULL){
//        while(buffer == ""){ m_srv_handler->receive(buffer); usleep(500); }
//        hex_string_print(buffer);
//      }
//      else{
//        std::cout << "m_srv_handler is NULL" << std::endl;
//        nullpo=true;
//      }
//    }
//    else{
//      if(m_cli_handler != NULL){
//        while(buffer == ""){ m_cli_handler->receive(buffer); usleep(500); }
//        hex_string_print(buffer);
//      }
//      else{
//        std::cout << "m_cli_handler is NULL" << std::endl;
//        nullpo=true;
//      }
//    }

    
    // stream buffer
    std::vector<std::string> stream_buffer;
    for(int i=0; i<16; i++){ stream_buffer.push_back(""); }
    std::vector<std::string>::iterator b_it=stream_buffer.begin();
    std::vector<std::string>::iterator c_it=stream_buffer.begin();
    int seek_point = 0;

    int peer_sz_char   = 0;
    int peer_sz_int    = 0;
    int peer_sz_float  = 0;
    int peer_sz_double = 0;
    int num_of_streams = 0;

    bool b_peer_little_endian = true;

    bool need_byte_swap = false;
    bool need_resize = false;
    if(host_sz_int != peer_sz_int){
      need_resize = true;
      if(b_host_little_endian != b_peer_little_endian){ need_byte_swap = true; }
    }
    else{
      if(b_host_little_endian != b_peer_little_endian){ need_byte_swap = true; }
    }

    int i_tmp;
    char *cpi=(reinterpret_cast<char*>(&i_tmp));
    float f_tmp;
    char *cpf=(reinterpret_cast<char*>(&f_tmp));
    double d_tmp;
    char *cpd=(reinterpret_cast<char*>(&d_tmp));
    int header_tmp[4];
    int stream_type[16];
    int stream_size[16];
    for(int i=0; i<16; i++){ stream_type[i]=0; stream_size[i]=0; }

    seek_point = 8;
    if(buffer.size() > seek_point+(8*peer_sz_int)){
      for(int i=0, offset=seek_point; i<8; i++, offset+=peer_sz_int){
        for(int j=0; j<peer_sz_int; j++){
          cpi[j] = (0xff & buffer.c_str()[offset+j]);
        }
        switch(i){
        //case 0:                         break; // reserved1
        //case 1:                         break; // version
        case 2: peer_sz_char   = i_tmp; break; // size_of_char
        case 3: peer_sz_int    = i_tmp; break; // size_of_int
        case 4: peer_sz_float  = i_tmp; break; // size_of_float
        case 5: peer_sz_double = i_tmp; break; // size_of_double
        //case 6:                         break; // reserved2
        case 7: num_of_streams = i_tmp; break; // number of streams
        default:                        break;
        }
        //std::cout << "SS " << setw(8) << setfill('0') << std::hex << i_tmp << std::endl;
      }
    }

    int end_mask = 0x10000010;
    seek_point += (8*peer_sz_int);

    for(int stream=0; stream<num_of_streams; stream++){
      for(int i=0, offset=seek_point; i<4; i++, offset+=peer_sz_int){
        for(int j=0; j<peer_sz_int; j++){
          cpi[j] = (0xff & buffer.c_str()[offset+j]);
        }
        header_tmp[i] = i_tmp;
      }
      //header_tmp[0]--;
      stream_type[header_tmp[0]] = header_tmp[1];
      if(header_tmp[1] == 0x01020F00){ // Source
        stream_size[header_tmp[0]] = (peer_sz_int * 4) + (header_tmp[2] * ((peer_sz_int) + (peer_sz_float * 4)));
        std::cout << " 0x01020F00 Source :: stream " << std::hex << header_tmp[0] << " size " << stream_size[header_tmp[0]] << std::endl;
      }
      else if(header_tmp[1] == 0x01020300){ // Vector<int>
        stream_size[header_tmp[0]] = (peer_sz_int * 4) + (header_tmp[2] * (peer_sz_int));
        std::cout << " 0x01020300 Vector<int> :: stream " << std::hex << header_tmp[0] << " size " << stream_size[header_tmp[0]] << std::endl;
      }
      else if(header_tmp[1] == 0x01020500){ // Vector<float>
        stream_size[header_tmp[0]] = (peer_sz_int * 4) + (header_tmp[2] * (peer_sz_float));
        std::cout << " 0x01020500 Vector<float> :: stream " << std::hex << header_tmp[0] << " size " << stream_size[header_tmp[0]] << std::endl;
      }
      else if(header_tmp[1] == 0x01020600){ // Vector<complex<float> >
        stream_size[header_tmp[0]] = (peer_sz_int * 4) + (header_tmp[2] * (peer_sz_float * 2));
        std::cout << " 0x01020600 Vector<complex<float> > :: stream " << std::hex << header_tmp[0] << " size " << stream_size[header_tmp[0]] << std::endl;
      }
      else if(header_tmp[1] == 0x01030300){ // Matrix<int>
        stream_size[header_tmp[0]] = (peer_sz_int * 5) + ((header_tmp[2] * header_tmp[3]) * (peer_sz_int));
        std::cout << " 0x01030300 Matrix<int> :: stream " << std::hex << header_tmp[0] << " size " << stream_size[header_tmp[0]] << std::endl;
      }
      else if(header_tmp[1] == 0x01030500){ // Matrix<float>
        stream_size[header_tmp[0]] = (peer_sz_int * 5) + ((header_tmp[2] * header_tmp[3]) * (peer_sz_float));
        std::cout << " 0x01030500 Matrix<float> :: stream " << std::hex << header_tmp[0] << " size " << stream_size[header_tmp[0]] << std::endl;
        std::cout << " header_tmp[0-3] " << std::hex << header_tmp[0] << " " << header_tmp[1] << " " << header_tmp[2] << " " << header_tmp[3] << std::endl;
        std::cout << " peer_sz_float " << std::hex << peer_sz_float << " peer_sz_int " << peer_sz_int << std::endl;
      }
      else if(header_tmp[1] == 0x01030600){ // Matrix<complex<float> >
        stream_size[header_tmp[0]] = (peer_sz_int * 5) + ((header_tmp[2] * header_tmp[3]) * (peer_sz_float * 2));
        std::cout << " 0x01020F00 Source :: stream " << std::hex << header_tmp[0] << " size " << stream_size[header_tmp[0]] << std::endl;
      }
      else if(header_tmp[1] == 0x01040F00){ // Map<int, ObjectRef> // UnSupported now. Map size only zero.
        stream_size[header_tmp[0]] = (peer_sz_int * 4);
        std::cout << " 0x01030600 Map<int, ObjectRef> :: stream " << std::hex << header_tmp[0] << " size " << stream_size[header_tmp[0]] << std::endl;
      }
      else if(header_tmp[1] == 0x01040500){ // Map<int, Vector<float> >
        int map_seek = (peer_sz_int*3); // Streams : StartHeader : MapSize
        for(int i=0; i<header_tmp[2]; i++){
          for(int j=0; j<peer_sz_int; j++){
            cpi[j] = (0xff & buffer.c_str()[seek_point+map_seek+peer_sz_int+j]);
          }
          map_seek += ((peer_sz_int * 2) + (i_tmp * peer_sz_float));
        }
        stream_size[header_tmp[0]] = (map_seek) + (peer_sz_int);
        std::cout << " 0x01040500 Map<int, Vector<float> > :: stream " << std::hex << header_tmp[0] << " size " << stream_size[header_tmp[0]] << std::endl;
      }
      else if(header_tmp[1] == 0x01040600){ // Map<int, Vector<complex<float> > >
        int map_seek = (peer_sz_int*3); // Streams : StartHeader : MapSize
        for(int i=0; i<header_tmp[2]; i++){
          for(int j=0; j<peer_sz_int; j++){
            cpi[j] = (0xff & buffer.substr(seek_point+map_seek+peer_sz_int, peer_sz_int).c_str()[j]);
          }
          map_seek += ((peer_sz_int * 2) + (i_tmp * peer_sz_float * 2));
        }
        stream_size[header_tmp[0]] = (map_seek) + (peer_sz_int);
        std::cout << " 0x01040600 Map<int, Vector<complex<float> > > :: stream " << std::hex << header_tmp[0] << " size " << stream_size[header_tmp[0]] << std::endl;
      }
      else if(header_tmp[1] == 0x01000000){ // No data
        stream_size[header_tmp[0]] = (peer_sz_int * 3);
        std::cout << " 0x01000000 No data :: stream " << std::hex << header_tmp[0] << " size " << stream_size[header_tmp[0]] << std::endl;
      }
      else{ // Unknown
        std::cout << " 0x???????? Unknown :: stream " << std::hex << header_tmp[0] << " size " << stream_size[header_tmp[0]] << std::endl;
      }
      //std::cout << "size " << stream_size[header_tmp[0]] << " stream " << header_tmp[0] << std::endl;

      stream_buffer[header_tmp[0]] = buffer.substr(seek_point, stream_size[header_tmp[0]]);
      seek_point += stream_size[header_tmp[0]];
    }

    for(int stream=0; stream<num_of_streams; stream++){
      if(stream_type[stream] == 0x01020F00){ // Source
        int offset=0;
        int size=0;
        Vector<ObjectRef>& output_vector_src_p = *new Vector<ObjectRef>; // ϥХåե
        (*(outputs[outputID[stream]].buffer))[count] = &output_vector_src_p;
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream number
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream start header
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } size = i_tmp; // Source size
        for(int j=0; j<size; j++){
          RCPtr<Source> src(new Source);
          for(int i=0; i<peer_sz_int; i++, offset++){   cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } src->id    = i_tmp;
          for(int i=0; i<peer_sz_float; i++, offset++){ cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); } src->x[0]  = f_tmp;
          for(int i=0; i<peer_sz_float; i++, offset++){ cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); } src->x[1]  = f_tmp;
          for(int i=0; i<peer_sz_float; i++, offset++){ cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); } src->x[2]  = f_tmp;
          for(int i=0; i<peer_sz_float; i++, offset++){ cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); } src->power = f_tmp;
          output_vector_src_p.push_back(src);
        }
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream end header
        std::cout << " 0x01020F00 Source :: converted." << std::endl;
      }
      else if(stream_type[stream] == 0x01020300){ // Vector<int>
        int offset=0;
        int vec_size=0;
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream number
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream start header
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } vec_size = i_tmp; // Vector size
        RCPtr<Vector<int> > output_vector_int_p(new Vector<int>(vec_size)); // ϥХåե
        (*(outputs[outputID[stream]].buffer))[count] = output_vector_int_p;
        Vector<int>& output = *output_vector_int_p;
        if(boutput_overlap[stream]){
          if(count == 0){
            //if(param_i_advance != vec_size){  } // (param_i_advance != vec_size) && (overlap_size == param_i_length - param_i_advance)
            for(int j=0; j<overlap_size; j++){ // out[0~(length-advance-1)] <== 0
              ////for(int i=0; i<peer_sz_int; i++, offset++){   cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output[j] = 0;
            }
            for(int j=overlap_size, k=0; j<param_i_length; j++, k++){ // out[(length-advance)~length-1] <== current[0~advance-1]
              for(int i=0; i<peer_sz_int; i++, offset++){   cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output[j] = i_tmp;
            }
          }
          else{
            Vector<int>& previous = object_cast<Vector<int> >((*(outputs[outputID[stream]].buffer))[count - 1]);
            //if(param_i_advance != vec_size){  } // (param_i_advance != vec_size) && (overlap_size == param_i_length - param_i_advance)
            for(int j=0, k=param_i_advance; j<overlap_size; j++, k++){ // out[0~(length-advance-1)] <== prev_buf[advance~length-1]
              ////for(int i=0; i<peer_sz_int; i++, offset++){   cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output[j] = previous[k];
            }
            for(int j=overlap_size, k=0; j<param_i_length; j++, k++){ // out[(length-advance)~length-1] <== current[0~advance-1]
              for(int i=0; i<peer_sz_int; i++, offset++){   cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output[j] = i_tmp;
            }
          }
        }
        else{ // no processing overlap
          for(int j=0; j<vec_size; j++){
            for(int i=0; i<peer_sz_int; i++, offset++){   cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
            output[j] = i_tmp;
          }
        }
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream end header
        std::cout << " 0x01020300 Vector<int> :: converted." << std::endl;
      }
      else if(stream_type[stream] == 0x01020500){ // Vector<float>
        int offset=0;
        int vec_size=0;
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream number
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream start header
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } vec_size = i_tmp; // Vector size
        RCPtr<Vector<float> > output_vector_float_p(new Vector<float>(vec_size)); // ϥХåե
        (*(outputs[outputID[stream]].buffer))[count] = output_vector_float_p;
        Vector<float>& output = *output_vector_float_p;
        if(boutput_overlap[stream]){
          if(count == 0){
            //if(param_i_advance != vec_size){  } // (param_i_advance != vec_size) && (overlap_size == param_i_length - param_i_advance)
            for(int j=0; j<overlap_size; j++){ // out[0~(length-advance-1)] <== 0.0f
              ////for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output[j] = 0.0f;
            }
            for(int j=overlap_size, k=0; j<param_i_length; k++, j++){ // out[(length-advance)~length-1] <== current[0~advance-1]
              for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output[j] = f_tmp;
            }
          }
          else{
            Vector<float>& previous = object_cast<Vector<float> >((*(outputs[outputID[stream]].buffer))[count - 1]);
            //if(param_i_advance != vec_size){  } // (param_i_advance != vec_size) && (overlap_size == param_i_length - param_i_advance)
            for(int j=0, k=param_i_advance; j<overlap_size; j++, k++){ // out[0~(length-advance-1)] <== prev_buf[advance~length-1]
              ////for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output[j] = previous[k];
            }
            for(int j=overlap_size, k=0; j<param_i_length; j++, k++){ // out[(length-advance)~length-1] <== current[0~advance-1]
              for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output[j] = f_tmp;
            }
          }
        }
        else{ // no processing overlap
          for(int j=0; j<vec_size; j++){
            for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
            output[j] = f_tmp;
          }
        }
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream end header
        std::cout << " 0x01020500 Vector<float> :: converted." << std::endl;
      }
      else if(stream_type[stream] == 0x01020600){ // Vector<complex<float> >
        int offset=0;
        int vec_size=0;
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream number
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream start header
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } vec_size = i_tmp; // Vector size
        RCPtr<Vector<complex<float> > > output_vector_complex_float_p(new Vector<complex<float> >(vec_size)); // ϥХåե
        (*(outputs[outputID[stream]].buffer))[count] = output_vector_complex_float_p;
        Vector<complex<float> >& output = *output_vector_complex_float_p;
        if(boutput_overlap[stream]){
          if(count == 0){
            if(param_i_advance != vec_size){  } // (param_i_advance != vec_size) && (overlap_size == param_i_length - param_i_advance)
            for(int j=0; j<overlap_size; j++){ // out[0~(length-advance-1)] <== 0.0f
              ////for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output[j].real() = 0.0f;
              ////for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output[j].imag() = 0.0f;
            }
            for(int j=overlap_size, k=0; j<param_i_length; j++, k++){ // out[(length-advance)~length-1] <== current[0~advance-1]
              for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output[j].real() = f_tmp;
              for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output[j].imag() = f_tmp;
            }
          }
          else{
            Vector<complex<float> >& previous = object_cast<Vector<complex<float> > >((*(outputs[outputID[stream]].buffer))[count - 1]);
            if(param_i_advance != vec_size){  } // (param_i_advance != vec_size) && (overlap_size == param_i_length - param_i_advance)
            for(int j=0, k=param_i_advance; j<overlap_size; j++, k++){ // out[0~(length-advance-1)] <== prev_buf[advance~length-1]
              ////for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output[j].real() = previous[j].real();
              ////for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output[j].imag() = previous[j].imag();
            }
            for(int j=overlap_size, k=0; j<param_i_length; j++, k++){ // out[(length-advance)~length-1] <== current[0~advance-1]
              for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output[j].real() = f_tmp;
              for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output[j].imag() = f_tmp;
            }
          }
        }
        else{ // no processing overlap
          for(int j=0; j<vec_size; j++){
            for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
            output[j].real() = f_tmp;
            for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
            output[j].imag() = f_tmp;
          }
        }
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream end header
        std::cout << " 0x01020600 Vector<complex<float> > :: converted." << std::endl;
      }
      else if(stream_type[stream] == 0x01030300){ // Matrix<int>
        int offset=0;
        int rows=0;
        int cols=0;
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream number
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream start header
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } rows = i_tmp; // Matrix row size
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } cols = i_tmp; // Matrix col size
        RCPtr<Matrix<int> > output_matrix_int_p(new Matrix<int>(rows, cols)); // ϥХåե
        (*(outputs[outputID[stream]].buffer))[count] = output_matrix_int_p;
        Matrix<int>& output = *output_matrix_int_p;
        if(boutput_overlap[stream]){
          if(count == 0){
            //if(param_i_advance != cols){  } // (param_i_advance != cols) && (overlap_size == param_i_length - param_i_advance)
            for(int j=0; j<rows; j++){ // rows is channels
              for(int k=0; k<overlap_size; k++){ // out(rows, 0~(length-advance-1)) <== 0.0f
                //for(int i=0; i<peer_sz_int; i++, offset++){   cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                output(j, k) = 0;
              }
              for(int k=overlap_size, l=0; k<param_i_length; k++, l++){ // out(rows, (length-advance)~length-1) <== current(rows, 0~advance-1)
                for(int i=0; i<peer_sz_int; i++, offset++){   cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                output(j, k) = i_tmp;
              }
            }
          }
          else{
            Matrix<int>& previous = object_cast<Matrix<int> >((*(outputs[outputID[stream]].buffer))[count - 1]);
            //if(param_i_advance != cols){  } // (param_i_advance != cols) && (overlap_size == param_i_length - param_i_advance)
            for(int j=0; j<rows; j++){ // rows is channels
              for(int k=0, l=param_i_advance; k<overlap_size; k++, l++){ // out(rows, 0~(length-advance-1)) <== prev_buf(rows, advance~length-1)
                //for(int i=0; i<peer_sz_int; i++, offset++){   cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                output(j, k) = previous(j, l);
              }
              for(int k=overlap_size, l=0; k<param_i_length; k++, l++){ // out(rows, (length-advance)~length-1) <== current(rows, 0~advance-1)
                for(int i=0; i<peer_sz_int; i++, offset++){   cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                output(j, k) = i_tmp;
              }
            }
          }
        }
        else{ // no processing overlap
          for(int j=0; j<rows; j++){ // rows is channels
            for(int k=0; k<cols; k++){ // cols is length
              for(int i=0; i<peer_sz_int; i++, offset++){   cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output(j, k) = i_tmp;
            }
          }
        }
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream end header
        std::cout << " 0x01030300 Matrix<int> :: converted." << std::endl;
      }
      else if(stream_type[stream] == 0x01030500){ // Matrix<float>
        int offset=0;
        int rows=0;
        int cols=0;
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream number
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream start header
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } rows = i_tmp; // Matrix row size
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } cols = i_tmp; // Matrix col size
        RCPtr<Matrix<float> > output_matrix_float_p(new Matrix<float>(rows, cols)); // ϥХåե
        (*(outputs[outputID[stream]].buffer))[count] = output_matrix_float_p;
        Matrix<float>& output = *output_matrix_float_p;
        if(boutput_overlap[stream]){
          if(count == 0){
            //if(param_i_advance != cols){  } // (param_i_advance != cols) && (overlap_size == param_i_length - param_i_advance)
            for(int j=0; j<rows; j++){ // rows is channels
              for(int k=0; k<overlap_size; k++){ // out(rows, 0~(length-advance-1)) <== 0.0f
                //for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                output(j, k) = 0.0f;
              }
              for(int k=overlap_size, l=0; k<param_i_length; k++, l++){ // out(rows, (length-advance)~length-1) <== current(rows, 0~advance-1)
                for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                output(j, k) = f_tmp;
              }
            }
          }
          else{
            Matrix<float>& previous = object_cast<Matrix<float> >((*(outputs[outputID[stream]].buffer))[count - 1]);
            //if(param_i_advance != cols){  } // (param_i_advance != cols) && (overlap_size == param_i_length - param_i_advance)
            for(int j=0; j<rows; j++){ // rows is channels
              for(int k=0, l=param_i_advance; k<overlap_size; k++, l++){ // out(rows, 0~(length-advance-1)) <== prev_buf(rows, advance~length-1)
                //for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                output(j, k) = previous(j, l);
              }
              for(int k=overlap_size, l=0; k<param_i_length; k++, l++){ // out(rows, (length-advance)~length-1) <== current(rows, 0~advance-1)
                for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                output(j, k) = f_tmp;
              }
            }
          }
        }
        else{ // no processing overlap
          for(int j=0; j<rows; j++){ // rows is channels
            for(int k=0; k<cols; k++){ // cols is length
              for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output(j, k) = f_tmp;
            }
          }
        }
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream end header
        std::cout << " 0x01030500 Matrix<float> :: converted." << std::endl;
      }
      else if(stream_type[stream] == 0x01030600){ // Matrix<complex<float> >
        int offset=0;
        int rows=0;
        int cols=0;
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream number
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream start header
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } rows = i_tmp; // Matrix row size
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } cols = i_tmp; // Matrix col size
        RCPtr<Matrix<complex<float> > > output_matrix_complex_float_p(new Matrix<complex<float> >(rows, cols)); // ϥХåե
        (*(outputs[outputID[stream]].buffer))[count] = output_matrix_complex_float_p;
        Matrix<complex<float> >& output = *output_matrix_complex_float_p;
        if(boutput_overlap[stream]){
          if(count == 0){
            //if(param_i_advance != cols){  } // (param_i_advance != cols) && (overlap_size == param_i_length - param_i_advance)
            for(int j=0; j<rows; j++){ // rows is channels
              for(int k=0; k<overlap_size; k++){ // out(rows, 0~(length-advance-1)) <== 0.0f
                //for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                output(j, k).real() = 0.0f;
                //for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                output(j, k).imag() = 0.0f;
              }
              for(int k=overlap_size, l=0; k<param_i_length; k++, l++){ // out(rows, (length-advance)~length-1) <== current(rows, 0~advance-1)
                for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                output(j, k).real() = f_tmp;
                for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                output(j, k).imag() = f_tmp;
              }
            }
          }
          else{
            Matrix<complex<float> >& previous = object_cast<Matrix<complex<float> > >((*(outputs[outputID[stream]].buffer))[count - 1]);
            //if(param_i_advance != cols){  } // (param_i_advance != cols) && (overlap_size == param_i_length - param_i_advance)
            for(int j=0; j<rows; j++){ // rows is channels
              for(int k=0, l=param_i_advance; k<overlap_size; k++, l++){ // out(rows, 0~(length-advance-1)) <== prev_buf(rows, advance~length-1)
                //for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                output(j, k).real() = previous(j, l).real();
                //for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                output(j, k).imag() = previous(j, l).imag();
              }
              for(int k=overlap_size, l=0; k<param_i_length; k++, l++){ // out(rows, (length-advance)~length-1) <== current(rows, 0~advance-1)
                for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                output(j, k).real() = f_tmp;
                for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                output(j, k).imag() = f_tmp;
              }
            }
          }
        }
        else{ // no processing overlap
          for(int j=0; j<rows; j++){ // rows is channels
            for(int k=0; k<cols; k++){ // cols is length
              for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output(j, k).real() = f_tmp;
              for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              output(j, k).imag() = f_tmp;
            }
          }
        }
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream end header
        std::cout << " 0x01030600 Matrix<complex<float> > :: converted." << std::endl;
      }
      else if(stream_type[stream] == 0x01040F00){ // Map<int, ObjectRef> // UnSupported now. Map size only zero.
        int offset=0;
        int size_map=0;
        int size_vec=0;
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream number
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream start header
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } size_map = i_tmp; // Map size
        // UnSupported yet. Always zero size map now.
        //for(int k=0; k<size_map; k++){
        //  for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // Map key ID
        //  for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } size_vec = i_tmp;
        //  for(int j=0; j<size_vec; j++){
        //    for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
        //  }
        //}
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream end header

        stream_size[header_tmp[0]] = (peer_sz_int * 4);
        std::cout << " 0x01030600 Map<int, ObjectRef> :: stream " << std::hex << header_tmp[0] << " size " << stream_size[header_tmp[0]] << std::endl;
      }
      else if(stream_type[stream] == 0x01040500){ // Map<int, Vector<float> >
        int offset=0;
        int map_size=0;
        int map_key=0;
        int vec_size=0;
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream number
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream start header
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } map_size = i_tmp; // Map size
        RCPtr<Map<int, ObjectRef> > output_map_p(new Map<int, ObjectRef>); // ϥХåե
        (*(outputs[outputID[stream]].buffer))[count]  = output_map_p;
        Map<int, ObjectRef>& output = *output_map_p;
        // if separated wave data, separated feature data or mfm data, Map<int, Vector<float> >
        // if separated fft data, Map<int, Vector<complex<float> > >
        for(int k=0; k<map_size; k++){ // map_size is sorces
          for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } map_key  = i_tmp; // Map key ID
          for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } vec_size = i_tmp;
          RCPtr<Vector<float> > vec_float(new Vector<float>(vec_size)); // vec_size is length
          if(boutput_overlap[stream]){
            if(count == 0){
              //if(param_i_advance != vec_size){  } // (param_i_advance != cols) && (overlap_size == param_i_length - param_i_advance)
              for(int j=0; j<overlap_size; j++){ // out[0~(length-advance-1)] <== 0
                //for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                (*vec_float)[j] = 0.0f;
              }
              for(int j=overlap_size, k=0; j<param_i_length; j++, k++){ // out[(length-advance)~length-1] <== current[0~advance-1]
                for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                (*vec_float)[j] = f_tmp;
              }
            }
            else{
              Map<int, ObjectRef>::iterator previous_it;
              Map<int, ObjectRef>& previous_map = object_cast<Map<int, ObjectRef> >((*(outputs[outputID[stream]].buffer))[count - 1]);
              for(Map<int, ObjectRef>::iterator it = previous_map.begin(); it != previous_map.end(); it++){
                if(it->first == map_key){ previous_it = it; }
              }
              Vector<float>& previous = object_cast<Vector<float> >(previous_it->second);
              //if(param_i_advance != vec_size){  } // (param_i_advance != vec_size) && (overlap_size == param_i_length - param_i_advance)
              for(int j=0, k=param_i_advance; j<overlap_size; j++, k++){ // out[0~(length-advance-1)] <== prev_buf[advance~length-1]
                ////for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                (*vec_float)[j] = previous[k];
              }
              for(int j=overlap_size, k=0; j<param_i_length; j++, k++){ // out[(length-advance)~length-1] <== current[0~advance-1]
                for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                (*vec_float)[j] = f_tmp;
              }
            }
          }
          else{ // no processing overlap
            for(int j=0; j<vec_size; j++){
              for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              (*vec_float)[j] = f_tmp;
            }
          }
          output[map_key] = vec_float; // map_key is sorce id
        }
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream end header
        std::cout << " 0x01040500 Map<int, Vector<float> > :: converted." << std::endl;
      }
      else if(stream_type[stream] == 0x01040600){ // Map<int, Vector<complex<float> > >
        int offset=0;
        int map_size=0;
        int map_key=0;
        int vec_size=0;
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream number
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream start header
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } map_size = i_tmp; // Map size
        RCPtr<Map<int, ObjectRef> > output_map_p(new Map<int, ObjectRef>); // ϥХåե
        (*(outputs[outputID[stream]].buffer))[count]  = output_map_p;
        Map<int, ObjectRef>& output = *output_map_p;
        // if separated wave data, separated feature data or mfm data, Map<int, Vector<float> >
        // if separated fft data, Map<int, Vector<complex<float> > >
        for(int k=0; k<map_size; k++){ // map_size is sorces
          for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } map_key  = i_tmp; // Map key ID
          for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } vec_size = i_tmp;
          RCPtr<Vector<complex<float> > > vec_complex_float(new Vector<complex<float> >(vec_size)); // vec_size is length
          if(boutput_overlap[stream]){
            if(count == 0){
              //if(param_i_advance != vec_size){  } // (param_i_advance != cols) && (overlap_size == param_i_length - param_i_advance)
              for(int j=0; j<overlap_size; j++){ // out[0~(length-advance-1)] <== 0
                //for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                (*vec_complex_float)[j].real() = 0.0f;
                //for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                (*vec_complex_float)[j].imag() = 0.0f;
              }
              for(int j=overlap_size, k=0; j<param_i_length; j++, k++){ // out[(length-advance)~length-1] <== current[0~advance-1]
                for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                (*vec_complex_float)[j].real() = f_tmp;
                for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                (*vec_complex_float)[j].imag() = f_tmp;
              }
            }
            else{
              Map<int, ObjectRef>::iterator previous_it;
              Map<int, ObjectRef>& previous_map = object_cast<Map<int, ObjectRef> >((*(outputs[outputID[stream]].buffer))[count - 1]);
              for(Map<int, ObjectRef>::iterator it = previous_map.begin(); it != previous_map.end(); it++){
                if(it->first == map_key){ previous_it = it; }
              }
              Vector<complex<float> >& previous = object_cast<Vector<complex<float> > >(previous_it->second);
              //if(param_i_advance != vec_size){  } // (param_i_advance != vec_size) && (overlap_size == param_i_length - param_i_advance)
              for(int j=0, k=param_i_advance; j<overlap_size; j++, k++){ // out[0~(length-advance-1)] <== prev_buf[advance~length-1]
                //for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                (*vec_complex_float)[j].real() = previous[k].real();
                //for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                (*vec_complex_float)[j].imag() = previous[k].imag();
              }
              for(int j=overlap_size, k=0; j<param_i_length; j++, k++){ // out[(length-advance)~length-1] <== current[0~advance-1]
                for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                (*vec_complex_float)[j].real() = f_tmp;
                for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
                (*vec_complex_float)[j].imag() = f_tmp;
              }
            }
          }
          else{ // no processing overlap
            for(int j=0; j<vec_size; j++){
              for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              (*vec_complex_float)[j].real() = f_tmp;
              for(int i=0; i<peer_sz_float; i++, offset++){   cpf[i] = (0xff & stream_buffer[stream].c_str()[offset]); }
              (*vec_complex_float)[j].imag() = f_tmp;
            }
          }
          output[map_key] = vec_complex_float; // map_key is sorce id
        }
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream end header
        std::cout << " 0x01040600 Map<int, Vector<complex<float> > > :: converted." << std::endl;
      }
      else if(stream_type[stream] == 0x01000000){ // No data
        int offset=0;
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream number
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream start header
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream end header

        std::cout << " 0x01000000 No data :: converted." << std::endl;
      }
      else{ // Unknown
        int offset=0;
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream number
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream start header
        for(int i=0; i<peer_sz_int; i++, offset++){ cpi[i] = (0xff & stream_buffer[stream].c_str()[offset]); } // stream end header

        std::cout << " 0x???????? Unknown :: converted." << std::endl;
      }
    }

/* for ROS code support (pending)
	for(int i = 0; i < cbf->deque_harkfeature[0].exist_src_num; i++){
	  RCPtr<Vector<float> > feature(new Vector<float>(cbf->deque_harkfeature[0].src[i].length));
	  for(int j = 0; j < cbf->deque_harkfeature[0].src[i].length; j++){
	    (*feature)[j] = cbf->deque_harkfeature[0].src[i].featuredata[j];
	  }
	  (*mapout)[cbf->deque_harkfeature[0].src[i].id] = feature;
	}
*/

/*

    Buffer &audioBuffer = *(outputs[audioID].buffer);
	RCPtr<Matrix<float> > outputp(new Matrix<float>(num_channels,length)); // ϥХåե
	audioBuffer[count] = outputp;
	Matrix<float>& output = *outputp;
        
	int readsize;
        
	/////////////////////////////////////////////
	// ǽΥ롼 (count == 0)
	/////////////////////////////////////////////
	if (count == 0) {
	  // ³Ԥ
	  cerr << "binding..." << endl;
	  if(bind(c_sock, (struct sockaddr *)&addr, sizeof(addr)) != 0){
		perror("binding error");
		exit(-1);
	  }
	  cerr << "listening..." << endl;
	  if (listen(c_sock, 1) != 0){
		perror("listen error");
		exit(-1);
	  }
	  
	  cerr << "accepting..." << endl;
	  socklen_t len = sizeof(client);
	  sock = accept(c_sock, (struct sockaddr *)&client, &len);
	  if (sock < 0){
		perror("accept error");
		exit(-1);
	  }

	  isConnected = 1;

	  // ϥХåեĴ
	  output.resize(num_channels, length);
          
	  // ꥵɤ߹
	  readsize = length * num_channels * sizeof(short int);
	  cerr << "receiving..." << endl;    
	  if (recv(sock, buffer, readsize, 0) == -1)
		eofBuffer[count] = FalseObject;
	  
	  // ϥХåե˽񤭹
	  cerr << "writing..." << endl;    
	  for (int r = 0; r < output.nrows(); r++) {
		for (int c = 0; c < output.ncols(); c++) {
		  output(r, c) = buffer[r + c * num_channels];
		}
	  }
	  cerr << "done." << endl;    
	}
	//////////////////////////////////////////////
	// ǥե (count >= 1)
	//////////////////////////////////////////////
	else {

	  if(!isConnected){
		cerr << "accepting... count:" << count;
		socklen_t len = sizeof(client);
		sock = accept(c_sock, (struct sockaddr *)&client, &len);
		cerr << "sock = " << sock;
		cerr << "  done." << endl;
		isConnected = 1;
	  }
	  
	  // Хåפ
	  if (advance < length) {
		Matrix<float>& previous =
		  object_cast<Matrix<float> >(audioBuffer[count - 1]);

		for (int c = 0; c < length - advance; c++) {
		  for (int r = 0; r < output.nrows(); r++) {
			output(r, c) = previous(r, c + advance);
		  }
		}

	  }
	  // ʤ: Τޤޥԡ
	  else {
		for (int c = 0; c < length - advance; c++) {
		  for (int r = 0; r < output.nrows(); r++) {
			output(r, c) = 0;
		  }
		}
	  }

	  // ꥵɤ߹
	  cerr << "recving...";
	  readsize = advance * num_channels * sizeof(short int) ;
	  if (recv(sock, buffer, readsize, MSG_WAITALL) == -1)
		eofBuffer[count] = TrueObject;//FalseObject;
	  cerr << "done" << eofBuffer[count] << endl;

	  int sum=0;
	  for(int tempi=0; tempi<advance*num_channels; tempi++){
		sum += buffer[tempi];
		cout << tempi << ":" << buffer[tempi] << ", ";
	  }
	  cout << endl;
	  cout << "sum: " << sum << endl;
	  if(sum == 0){
		isConnected = 0;
		return;
	  }

	  // ϥХåե˽񤭹
	  for (int r = 0; r < output.nrows(); r++) {
		for (int c = length - advance; c < output.ncols(); c++) {
		  output(r, c) = buffer[r + (c - length + advance) * num_channels];
		}
	  }

	}

*/

  }

/*
  int readfull(int id, void *buf, unsigned int size) {
	int ret;
	unsigned int stack = 0, diff = size;
	unsigned char *pbuf = (unsigned char *)buf;
        
	while (stack < size) {
	  if ((ret = read(id, &pbuf[stack], diff)) <= 0)
		return stack;
	  stack += ret;
	  diff  -= ret;
	}
        
	return stack;
  }

  ///////////////////////////////////////////////////////////////
  // WebSocket server start
  ///////////////////////////////////////////////////////////////
  int startServer(int port, bool use_limit, int limit){
    int status_code = 0;
    try{
      server::handler::ptr handler(new ws_server_handler()); // create an instance of handler
      server endpoint(handler);                              // create a server the listen on port uses handler
      // set log level for message print
      endpoint.alog().set_level(websocketpp::log::alevel::CONNECT);
      endpoint.alog().set_level(websocketpp::log::alevel::DISCONNECT);
      endpoint.elog().set_level(websocketpp::log::elevel::RERROR);
      endpoint.elog().set_level(websocketpp::log::elevel::FATAL);
      if(use_limit){
        server->set_max_messages_size(limit);
        cout << "Set receive message limitation [" << limit << " Byte]." << endl;
      }
      cout << "Starting WebSocket server on port [" << port << "]." << endl;
      endpoint.listen(port);
      status_code = 1;
    }
    catch(exception& e){
      cerr << "Exception: " << e.what() << endl;
      status_code = -1;
    }
    return status_code;
  }


  ///////////////////////////////////////////////////////////////
  // WebSocket client start
  ///////////////////////////////////////////////////////////////
  int startClient(int port, bool use_limit, int limit){
    int status_code = 0;
    string uri = "ws://localhost:8003/hark";
    try{
      //ws_client_handler_ptr handler(new chat_client_handler());
      client::handler::ptr handler(new ws_client_handler()); // create an instance of handler
      client endpoint(handler);                              // create a client uses handler
      client::connection_ptr connect_ptr;
      // set log level for message print
      endpoint.alog().unset_level(websocketpp::log::alevel::ALL); // disable all infomation message
      endpoint.elog().unset_level(websocketpp::log::elevel::ALL); // disable all error message
      endpoint.elog().set_level(websocketpp::log::elevel::RERROR);
      endpoint.elog().set_level(websocketpp::log::elevel::FATAL);
      // get connection
      connect_ptr = endpoint.get_connection(uri);
      connect_ptr->add_request_header("User-Agent", "WebSocket++/0.2.0 FlowDesignerSubscriberWebSocket/0.0.1");
      connect_ptr->add_subprotocol("hri.websocketpp.hark");
      connect_ptr->set_origin("http://hark.jp.honda-ri.com");
      endpoint.connect(connect_ptr);
      boost::thread thread_hark(boost::bind(&client::run, &endpoint, false));

      char test_buffer[2048];
      while(cin.getline(test_buffer, 2048)){ handler->send(test_buffer); }

      thread_hark.join();
      cout << "Starting WebSocket client." << endl;
      status_code = 1;
    }
    catch(exception& e){
      cerr << "Exception: " << e.what() << endl;
      status_code = -1;
    }
    return status_code; // 1:success, 0:throw, -1:exception
  }

	///////////////////////////////////////////////////////////////
    //
    // connect to a socket server
    //
    ///////////////////////////////////////////////////////////////
    int connectToServer() {
	int socket_id;
	
	hostent* connect_host;
	if ((connect_host = gethostbyname(hostname.c_str())) == NULL) {
	    cerr << "cannot find the host." << endl;
	    return -3;
	}
	
	sockaddr_in desthost;
	bzero((char*)&desthost, sizeof(desthost));
	desthost.sin_family = AF_INET;
	desthost.sin_port = htons(port);
	bcopy(connect_host->h_addr, (char*)&desthost.sin_addr, connect_host->h_length);
	
	if ((socket_id = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	    cerr << "cannot get socket." << endl;
	    return -2;
	}
	
	if (connect(socket_id, (sockaddr*)&desthost, sizeof(desthost)) == -1) {
	    cerr << "connection refused." << endl;
	    close(socket_id);
	    return -1;
	}
	
	return socket_id;
    }
*/
  void hex_string_print(std::string print_msg){
    /*
    for(int i=0; i<print_msg.size(); i++){
      int tc = 0xff & print_msg.c_str()[i];
      cout << hex << "0x" << setw(2) << setfill('0') << internal << tc << " ";
    }
    cout << dec << endl;
    */
  }

  IN_ORDER_NODE_SPEEDUP(SubscriberWebSocket)
};

//#endif // ENABLE_WEBSOCKET

