/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

//#ifndef HARK_MQTT_BASE_HPP
//#define HARK_MQTT_BASE_HPP

#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <queue>

#include <iomanip>

#include <cstring>

#include <mosquittopp.h>
#include "HARK_MQTT.hpp"

namespace HARK_MQTT {
  //-------- Publisher --------//
  // setter
  bool mqtt_publisher::send_message(std::string _topic, std::string _message, int _retry, int _qos, int _count){
    int ret, i;
    // In case of binary
    //if(is_debug){ std::cerr << "Publishing:" << std::endl << "  Topic=[" << (_topic.length()==0 ? topic : _topic) << "]" << std::endl << "  PayloadLength=" << _message.length() << std::endl << "  Payload=[" << _message << "]" << std::endl; }
    i = _retry;
    if(_topic.length()==0){
      do{
        ret = publish(/*mid=*/NULL, topic.c_str(), /*payloadlen=*/_message.length(), /*payload=*/_message.c_str(), _qos, /*retain=*/false);
        i--;
      } while(ret != MOSQ_ERR_SUCCESS && i >= 0);
    }
    else{
      do{
        ret = publish(/*mid=*/NULL, _topic.c_str(), /*payloadlen=*/_message.length(), /*payload=*/_message.c_str(), _qos, /*retain=*/false);
        i--;
      } while(ret != MOSQ_ERR_SUCCESS && i >= 0);
    }
    life += (ret == MOSQ_ERR_SUCCESS) ? 1 : 0;
    return ret == MOSQ_ERR_SUCCESS;
  }
  
  // for regacy implement compatibility
  bool mqtt_publisher::send_message(std::string _message, int _qos, int _count){
    //return publish("", _message, 0, _qos, _count);
    return send_message(topic, _message, 0, _qos, _count);
  }
  
  bool mqtt_publisher::send_message(char *_topic, int _topic_length, char *_message, int _message_length, int _retry, int _qos, int _count){
    //int _message_length = sizeof(_message)/_message[0];
    int ret, i;
    char *t, *p;
    p = new char[_message_length+1]; p[_message_length] = 0; // \0 termination for safe cord confirmation.
    std::copy(_message, _message + _message_length, p);
    if(_topic==NULL){
      // In case if binary
      //if(is_debug){ std::cerr << "Publishing:" << std::endl << "  Topic=[" << topic.c_str() << "]" << std::endl << "  PayloadLength=" << length << std::endl << "  Payload=[" << p << "]" << std::endl; }
      i = _retry;
      do{
        ret = publish(/*mid=*/NULL, topic.c_str(), /*payloadlen=*/_message_length, /*payload=*/p, _qos, /*retain=*/false);
        i--;
      } while(ret != MOSQ_ERR_SUCCESS && i >= 0);
    }
    else{
      t = new char[_topic_length+1]; t[_topic_length] = 0;
      std::copy(_topic, _topic + _topic_length, t);
      // In case if binary
      //if(is_debug){ std::cerr << "Publishing:" << std::endl << "  Topic=[" << t << "]" << std::endl << "  PayloadLength=" << length << std::endl << "  Payload=[" << p << "]" << std::endl; }
      i = _retry;
      do{
        ret = publish(/*mid=*/NULL, t, /*payloadlen=*/_message_length, /*payload=*/p, _qos, /*retain=*/false);
        i--;
      } while(ret != MOSQ_ERR_SUCCESS && i >= 0);
      delete[] t;
    }
    delete[] p;
    life += (ret == MOSQ_ERR_SUCCESS) ? 1 : 0;
    return ret == MOSQ_ERR_SUCCESS;
  }
  
  // for regacy implement compatibility
  bool mqtt_publisher::send_message(char *_message, int _message_length, int _qos, int _count){
    return send_message(NULL, 0, _message, _message_length, 0, _qos, _count);
  }
  
  int mqtt_publisher::get_life(){
    return life;
  }
  
  /*
  void mqtt_publisher::request_loop(){
    loop(100);
  }
  */
  
  // callbacks
  void mqtt_publisher::on_publish(int mid){
    life--;
    if(is_debug){ std::cerr << "Publish succeeded (" << mid << ")" << std::endl; }
  }
    
  void mqtt_publisher::on_connect(int rc){
    if(is_debug){
      switch(rc){
      case 0:
        std::cerr << "Connection suceeded (rc=0)" << std::endl;
        break;
      case 1:
        std::cerr << "Connection refused [unacceptable protocol] (rc=1)" << std::endl;
        break;
      case 2:
        std::cerr << "Connection refused [identifier rejected] (rc=2)" << std::endl;
        break;
      case 3:
        std::cerr << "Connection refused [broker unavailable] (rc=3)" << std::endl;
        break;
      default:
        std::cerr << "Connection refused [unknown (reserved value)] (rc=" << rc << ")" << std::endl;
        break;
      }
    }
  }
  
  void mqtt_publisher::on_disconnect(int rc) {
    if(is_debug){
      switch(rc){
      case 0:
        std::cerr << "Disconnection suceeded (rc=0)" << std::endl;
        break;
      default:
        std::cerr << "Disconnection failed [unknown (reserved value)] (rc=" << rc << ")" << std::endl;
        break;
      }
    }
  }
  
  //-------- Subscriber --------//
  // getter
  std::string mqtt_subscriber::receive_message(){
    if(is_debug){ std::cerr << "mqtt_subscriber::receive_message"  << std::endl; }

    std::string single_message = ""; // non blocking method, if buffer is empty, return string is "".
    if(!message_buffer.empty()){
      single_message = message_buffer.front();
      message_buffer.pop();
    }
    return single_message;
  }
    
  const char * mqtt_subscriber::receive_char_message(){
    if(is_debug){ std::cerr << "mqtt_subscriber::receive_char_message"  << std::endl; }

    std::string single_message("");
    if(!message_buffer.empty()){
      single_message = message_buffer.front();
      message_buffer.pop();
    }
    return single_message.c_str();
  }
    
  // callbacks
  void mqtt_subscriber::on_connect(int rc){
    if(is_debug){ std::cerr << "mqtt_subscriber::on_connect"  << std::endl; }

    if(is_debug){
      switch(rc){
      case 0:
        std::cerr << "Connection suceeded (rc=0)" << std::endl;
        break;
      case 1:
        std::cerr << "Connection refused [unacceptable protocol] (rc=1)" << std::endl;
        break;
      case 2:
        std::cerr << "Connection refused [identifier rejected] (rc=2)" << std::endl;
        break;
      case 3:
        std::cerr << "Connection refused [broker unavailable] (rc=3)" << std::endl;
        break;
      default:
        std::cerr << "Connection refused [unknown (reserved value)] (rc=" << rc << ")" << std::endl;
        break;
      }
    }

    if(rc == 0){
      connect_valid = true;

      if(is_debug){ std::cerr << "subscribing topic " << topic << std::endl; }
      subscribe(NULL, topic.c_str());
    }
    else connect_valid = false;
  }
  
  void mqtt_subscriber::on_disconnect(int rc) {
    if(is_debug){ std::cerr << "mqtt_subscriber::on_disconnect"  << std::endl; }

    if(is_debug){
      switch(rc){
      case 0:
        std::cerr << "Disconnection suceeded (rc=0)" << std::endl;
        break;
      default:
        std::cerr << "Disconnection failed [unknown (reserved value)] (rc=" << rc << ")" << std::endl;
        break;
      }
    }

    connect_valid = false;
  }
    
  void mqtt_subscriber::on_message(const struct mosquitto_message *msg){
    if(is_debug){ std::cerr << "mqtt_subscriber::on_message"  << std::endl; }

    int rc;
    if(is_debug){ std::cerr << "on message to " << msg->topic << std::endl; }
    if(!strcmp(msg->topic, topic.c_str())){
      //std::string tmp_str = (char*) msg->payload; // bad code
      std::string tmp_str; tmp_str.assign((char*) msg->payload, msg->payloadlen);
      message_buffer.push(tmp_str);
      // In case of string print
      //if(is_debug){
      //  std::cout << "received message: " << tmp_str << std::endl;
      //}
      // In case of binary dump
      //if(is_debug){
      //  std::cout << "received message:" << std::endl;
      //  for(std::string::iterator it=tmp_str.begin(); it<tmp_str.end(); it++){
      //    std::cerr << std::hex << std::setw(2) << ((int) *it & 0xff) << " ";
      //  }
      //  std::cerr << std::dec << std::endl;
      //}
    }
    // replace from loop() to loop_start(). I think no problem on subscriber.
    //rc = loop();
    //if(rc){ reconnect(); }
  }
  
  void mqtt_subscriber::on_subscribe(int mid, int qos_count, const int *granted_qos){
    if(is_debug){ std::cerr << "subscription started:" << qos_count << " " << *granted_qos << std::endl; }

  }
} // namespace HARK_MQTT end

//#endif // HARK_MQTT_BASE_HPP

