/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

//#ifdef ENABLE_MQTT

#include <unistd.h>

#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <vector>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Map.h>
#include <Source.h>

//#include <mosquittopp.h>
#define PUBLISHER 1 // bad work, will fix later.
#include "HARK_MQTT.hpp"
#include "HARK_FLAC.hpp"
#include "HARK_UUID.hpp"

using namespace FD;
using namespace HARK_MQTT;
//using namespace HARK_FLAC;
using namespace HARKPF::HARK_FLAC; // bad work, will fix later.

class PublisherFLAC;

DECLARE_NODE(PublisherFLAC);
/*Node
 *
 * @name PublisherFLAC
 * @category HARK:FLAC
 * @description Publish FLAC encode data via MQTT
 *
 * @input_name IN_SOURCES
 * @input_type Vector<ObjectRef>
 * @input_description Source localization results. Each element of the vector is Source type.
 *
 * @input_name IN_PCM
 * @input_type any
 * @input_description Input PCM data. Each element of the Matrix/Map is PCM data type.
 *
 * @output_name OUTPUT
 * @output_type Vector<ObjectRef>
 * @output_description The same as the output.
 *
 * @parameter_name ENABLE_SAVE
 * @parameter_type bool
 * @parameter_value false
 * @parameter_description Debug option. If true, output flac files.
 *
 * @parameter_name SAVE_FILENAME_PREFIX
 * @parameter_type string
 * @parameter_value published_
 * @parameter_description Duplicate from published flac stream per each Source ID to files.[default: published_]
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The shift length between adjacent frames (in samples) [default: 160].
 *
 * @parameter_name FLAC_CHANNEL
 * @parameter_type int
 * @parameter_value 8
 * @parameter_list 1:2:3:4:5:6:7:8
 * @parameter_description The number of microphone [default: 8]. Support channels listed up from FLAC library API's specification.
 *
 * @parameter_name FLAC_SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_list 88200:176400:192000:8000:16000:22050:24000:32000:44100:48000:96000
 * @parameter_description Sampling rate (Hz) [default: 16000]. Support sampling-rate listed up from FLAC library API's specification.
 *
 * @parameter_name FLAC_BITS_PER_SAMPLE
 * @parameter_type int
 * @parameter_value 24
 * @parameter_list 8:12:16:20:24
 * @parameter_description Selector of FLAC's bit per sample. FLAC default is 16[bits], but HARK default is 24[bits] because HARK is using by many 24 bits microphone.
 *
 * @parameter_name FLAC_COMPRESSION_LEVEL
 * @parameter_type int
 * @parameter_value 5
 * @parameter_list 0:1:2:3:4:5:6:7:8:15
 * @parameter_description FLAC compression level 0-8, reserved level 9-14 and manual setting mode (=15). Recommended strongly that leave the 5 [default: 5].
 *
 * @parameter_name FLAC_DO_MID_SIDE_STEREO
 * @parameter_type bool
 * @parameter_value true
 * @parameter_list true:false
 * @parameter_valid FLAC_COMPRESSION_LEVEL=15
 * @parameter_description Set to true to enable mid-side encoding on stereo input. The number of channels must be 2 for this to have any effect. Set to false to use only independent channel coding. Default value as same as FLAC Compression level 5 [default: true].
 *
 * @parameter_name FLAC_LOOSE_MID_SIDE_STEREO
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_valid FLAC_COMPRESSION_LEVEL=15
 * @parameter_description Set to true to enable adaptive switching between mid-side and left-right encoding on stereo input. Set to false to use exhaustive searching. Setting this to true requires FLAC_DO_MID_SIDE_STEREO to also be set to true in order to have any effect. Default value as same as FLAC Compression level 5 [default: false].
 *
 * @parameter_name FLAC_APODIZATION
 * @parameter_type string
 * @parameter_value tukey(0.5)
 * @parameter_list bartlett:bartlett_hann:blackman:blackman_harris_4term_92db:connes:flattop:gauss(0.1):gauss(0.2):gauss(0.3):gauss(0.4):gauss(0.5):hamming:hann:kaiser_bessel:nuttall:rectangle:triangle:tukey(0.125):tukey(0.25):tukey(0.5):welch
 * @parameter_valid FLAC_COMPRESSION_LEVEL=15
 * @parameter_description Sets the apodization function(s) the encoder will use when windowing audio data for LPC analysis. Note: tukey(0.0)==rectangle, tukey(1.0)==hann. Default value as same as FLAC Compression level 5 [default: tukey(0.5)].
 *
 * @parameter_name FLAC_MAX_LPC_ORDER
 * @parameter_type int
 * @parameter_value 8
 * @parameter_valid FLAC_COMPRESSION_LEVEL=15
 * @parameter_description Set the maximum LPC order, or 0 to use only the fixed predictors. Default value as same as FLAC Compression level 5 [default: 8].
 *
 * @parameter_name FLAC_QLP_COEFF_PRECISION
 * @parameter_type int
 * @parameter_value 0
 * @parameter_list 0:1:2:3:4:5:6:7:8:9:10:11:12:13:14:15:16:17:18:19:20:21:22:23
 * @parameter_valid FLAC_COMPRESSION_LEVEL=15
 * @parameter_description Set the precision, in bits. of the quantized linear predictor coefficients. If set to 0, to let the encoder select it based on the blocksize. In the current implementation, FLAC_QLP_COEFF_PRECISION + FLAC_BIT_PER_SAMPLE must be less than 32. Default value as same as FLAC Compression level 5 [default: 0].
 *
 * @parameter_name FLAC_DO_QLP_COEFF_PREC_SEARCH
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_valid FLAC_COMPRESSION_LEVEL=15
 * @parameter_description Set to false to use only the specified quantized linear predictor coefficient precision, or true to search neighboring precision values and use the best one. Default value as same as FLAC Compression level 5 [default: false].
 *
 * @parameter_name FLAC_DO_ESCAPE_CODING
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list false
 * @parameter_valid FLAC_COMPRESSION_LEVEL=15
 * @parameter_description Deprecated. Setting this value has no effect.
 *
 * @parameter_name FLAC_DO_EXHAUSTIVE_MODEL_SEARCH
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_valid FLAC_COMPRESSION_LEVEL=15
 * @parameter_description Set to false to let the encoder estimate the best model order based on the residual signal energy, or true to force the encoder to evaluate all order models and select the best. However, true is very heavy. Default value as same as FLAC Compression level 5 [default: false].
 *
 * @parameter_name FLAC_MIN_RESIDUAL_PARTITION_ORDER
 * @parameter_type int
 * @parameter_value 0
 * @parameter_list 0:1:2:3:4:5:6:7:8:9:10:11:12:13:14:15:16
 * @parameter_valid FLAC_COMPRESSION_LEVEL=15
 * @parameter_description Set the minimum partition order to search when coding the residual. The partition order determines the context size in the residual. The context size will be approximately "blocksize / (2 ** order)". Default value as same as FLAC Compression level 5 [default: 0].
 *
 * @parameter_name FLAC_MAX_RESIDUAL_PARTITION_ORDER
 * @parameter_type int
 * @parameter_value 5
 * @parameter_list 0:1:2:3:4:5:6:7:8:9:10:11:12:13:14:15:16
 * @parameter_valid FLAC_COMPRESSION_LEVEL=15
 * @parameter_description Set the maximum partition order to search when coding the residual. The partition order determines the context size in the residual. The context size will be approximately "blocksize / (2 ** order)". Default value as same as FLAC Compression level 5 [default: 5].
 *
 * @parameter_name MQTT_HOST
 * @parameter_type string
 * @parameter_value localhost
 * @parameter_description The host name of MQTT broker
 *
 * @parameter_name MQTT_PORT
 * @parameter_type int
 * @parameter_value 1883
 * @parameter_description The port of MQTT broker
 *
 * @parameter_name MQTT_TOPIC
 * @parameter_type string
 * @parameter_value hark/data/flac
 * @parameter_description The topic name to send the source
 *
 * @parameter_name MQTT_USERNAME
 * @parameter_type string
 * @parameter_value 
 * @parameter_description The username name to auth MQTT
 *
 * @parameter_name MQTT_PASSWORD
 * @parameter_type string
 * @parameter_value 
 * @parameter_description The password to auth MQTT
 *
 * @parameter_name MQTT_CAFILE
 * @parameter_type string
 * @parameter_value 
 * @parameter_description The ca file to SSL/TSL of MQTT
 *
 * @parameter_name MQTT_TIMEOUT
 * @parameter_type int
 * @parameter_value 60
 * @parameter_description The time out between MQTT messages
 *
 * @parameter_name MQTT_QoS
 * @parameter_type int
 * @parameter_list 0:1:2
 * @parameter_value 2
 * @parameter_description Quality of Service. The message is sent at most once if 0, at least once if 1, and exactly once if 2.
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_value false
 * @parameter_description Debug option. If true, it prints input list and output list.
 *
 END*/

/*


*/

class PublisherFLAC : public BufferedNode {
  bool param_b_enable_save;
  bool param_b_debug;
  bool param_b_manual_setting;
  bool param_b_do_mid_side_stereo;
  bool param_b_loose_mid_side_stereo;
  bool param_b_do_qlp_coeff_prec_search;
  bool param_b_do_escape_coding;
  bool param_b_do_exhaustive_model_search;
  mqtt_publisher *mqttobj;
  int input_sourcesID;
  int input_pcmID;
  int outputID;
  int param_i_advance;
  int param_i_channel;
  int param_i_sampling_rate;
  int param_i_bits_per_sample;
  int param_i_compression_level;
  int param_i_max_lpc_order;
  int param_i_qlp_coeff_precision;
  int param_i_min_residual_partition_order;
  int param_i_max_residual_partition_order;
  int param_i_keepalive;
  int param_i_timeout;
  int param_i_qos;
  int prev_life;
  //bool isDebug;
  std::string param_s_filename_prefix;
  std::string param_s_apodization;
  std::map<int, std::map<std::string, std::string> > sources_map;
  std::map<int, RequestServer::EncodeServer *> servers_map;
public:
  PublisherFLAC(std::string nodeName, ParameterSet params) : BufferedNode(nodeName, params), input_sourcesID(-1)
  {
    prev_life = 0;
    //input_sourcesID = addInput("IN_SOURCES");
    input_pcmID = addInput("IN_PCM");
    outputID = addOutput("OUTPUT");
    
    param_b_enable_save          = dereference_cast<bool>(parameters.get("ENABLE_SAVE"));
    param_s_filename_prefix      = object_cast<std::string>(parameters.get("SAVE_FILENAME_PREFIX"));
    param_i_advance              = dereference_cast<int>(parameters.get("ADVANCE"));

    // FLAC encoder parameters
    param_i_channel              = dereference_cast<int>(parameters.get("FLAC_CHANNEL"));
    param_i_sampling_rate        = dereference_cast<int>(parameters.get("FLAC_SAMPLING_RATE"));
    param_i_bits_per_sample      = dereference_cast<int>(parameters.get("FLAC_BITS_PER_SAMPLE"));
    param_i_compression_level    = dereference_cast<int>(parameters.get("FLAC_COMPRESSION_LEVEL"));
    if(param_i_compression_level>8){ // manual setting mode
      param_i_compression_level = 5; // based on compression level 5.
      param_b_manual_setting = true;
      param_b_do_mid_side_stereo           = dereference_cast<bool>(parameters.get("FLAC_DO_MID_SIDE_STEREO"));
      param_b_loose_mid_side_stereo        = dereference_cast<bool>(parameters.get("FLAC_LOOSE_MID_SIDE_STEREO"));
      param_s_apodization                  = object_cast<std::string>(parameters.get("FLAC_APODIZATION"));
      param_i_max_lpc_order                = dereference_cast<int>(parameters.get("FLAC_MAX_LPC_ORDER"));
      param_i_qlp_coeff_precision          = dereference_cast<int>(parameters.get("FLAC_QLP_COEFF_PRECISION"));
      param_b_do_qlp_coeff_prec_search     = dereference_cast<bool>(parameters.get("FLAC_DO_QLP_COEFF_PREC_SEARCH"));
      param_b_do_escape_coding             = dereference_cast<bool>(parameters.get("FLAC_DO_ESCAPE_CODING"));
      param_b_do_exhaustive_model_search   = dereference_cast<bool>(parameters.get("FLAC_DO_EXHAUSTIVE_MODEL_SEARCH"));
      param_i_min_residual_partition_order = dereference_cast<int>(parameters.get("FLAC_MIN_RESIDUAL_PARTITION_ORDER"));
      param_i_max_residual_partition_order = dereference_cast<int>(parameters.get("FLAC_MAX_RESIDUAL_PARTITION_ORDER"));
    }
    else{ // set compression level 5 parameters, but not used. effect of the compression level is applied only.
      param_b_manual_setting               = false;
      param_b_do_mid_side_stereo           = true;
      param_b_loose_mid_side_stereo        = false;
      param_s_apodization                  = "tukey(0.5)";
      param_i_max_lpc_order                = 8;
      param_i_qlp_coeff_precision          = 0;
      param_b_do_qlp_coeff_prec_search     = false;
      param_b_do_escape_coding             = false;
      param_b_do_exhaustive_model_search   = false;
      param_i_min_residual_partition_order = 0;
      param_i_max_residual_partition_order = 5;
    }

    // MQTT parameters
    std::string param_s_hostname = object_cast<std::string>(parameters.get("MQTT_HOST"));
    std::string param_s_topic    = object_cast<std::string>(parameters.get("MQTT_TOPIC"));
    int param_i_port             = dereference_cast<int>(parameters.get("MQTT_PORT"));
    std::string param_s_username, param_s_password, param_s_cafile;
    if(parameters.exist("MQTT_USERNAME")){
        param_s_username = object_cast<std::string>(parameters.get("MQTT_USERNAME"));
    }
    if(parameters.exist("MQTT_PASSWORD")){
        param_s_password = object_cast<std::string>(parameters.get("MQTT_PASSWORD"));
    }
    if(parameters.exist("MQTT_CAFILE")){
        param_s_cafile = object_cast<std::string>(parameters.get("MQTT_CAFILE"));
    }
    param_i_qos                  = dereference_cast<int>(parameters.get("MQTT_QoS"));
    
    param_b_debug                = dereference_cast<bool>(parameters.get("DEBUG"));
    param_i_keepalive            = 60;
    param_i_timeout              = dereference_cast<int>(parameters.get("MQTT_TIMEOUT"));
    
    std::cerr << param_s_username << " : " << param_s_password << std::endl;

    ////interval_count = interval;
    //int keepalive = 60;
    //if(param_b_debug){
    //  std::cerr << "HOST  : " << param_s_hostname << std::endl;
    //  std::cerr << "PORT  : " << param_i_port << std::endl;
    //  std::cerr << "TOPIC : " << param_s_topic << std::endl;
    //  std::cerr << "QOS   : " << param_i_qos << std::endl;
    //}
    //mqttobj = new mqtt_publisher(/*NULL*/"PUBLISHER", param_s_topic, param_s_hostname, param_i_port, keepalive);
    //mqttobj = new mqtt_publisher(/*NULL*/"PUBLISHER", param_s_topic, param_s_hostname, param_i_port, param_i_keepalive);
    //mqttobj = new mqtt_publisher(/*NULL*/"PUBLISHER", param_s_topic, param_s_hostname, param_i_port,
    //    param_s_username, param_s_password, param_s_cafile, param_i_keepalive, true, param_b_debug);
    HARK_UUID::uuid uuid;
    std::string unique_str = uuid.v4();
    std::string publisher_id("HARK|PublusherFLAC|" + unique_str);
    mqttobj = new mqtt_publisher(publisher_id.c_str(), param_s_topic, param_s_hostname, param_i_port,
        param_s_username, param_s_password, param_s_cafile, param_i_keepalive, true, param_b_debug);

    if(param_b_debug){
      std::cerr << "HOST  : " << param_s_hostname << std::endl;
      std::cerr << "PORT  : " << param_i_port << std::endl;
      std::cerr << "TOPIC : " << param_s_topic << std::endl;
      std::cerr << "QOS   : " << param_i_qos << std::endl;
    }
    
    inOrder = true;
  }

  ~PublisherFLAC(){
    if(param_b_debug){ std::cerr << "Destructing PublisherFLAC" << std::endl; }
    std::map<int, RequestServer::EncodeServer *>::iterator itenc;
    for(itenc=servers_map.begin(); itenc!=servers_map.end();){
      itenc->second->stop(); // Stop server

      // Send remaining flac data
      std::vector<unsigned char> charbuf = itenc->second->get_encoded_flac(65536, true);
      std::string dat; dat.assign(charbuf.begin(), charbuf.end());
      mqttobj->send_message(dat, param_i_qos, /*count=*/0); // count argument was not effective now.
      if(param_b_enable_save){
        std::ostringstream filename_str;
        std::ofstream ofs;
        filename_str << param_s_filename_prefix << ".flac";
        ofs.open(filename_str.str().c_str(), std::ios::out|std::ios::binary|std::ios::app);
        if(ofs){
          ofs.write((char*)&charbuf[0], charbuf.size());
          ofs.close();
        }
      }

      delete itenc->second; // Delete instance
      servers_map.erase(itenc++); // Erase element on map
    }
    //servers_map.clear(); // for debug
    std::cerr << "Publishing remaining message.";
    mqttobj->send_message("STOP_STREAM", param_i_qos, /*count=*/0); // count argument was not effective now.
    while(mqttobj->get_life()>0){ /*mqttobj->request_loop();*/ std::cerr << "."; } std::cerr << " finished." << std::endl;
    delete mqttobj;
    if(param_b_debug){ std::cerr << "Destructing Finished" << std::endl; }
  }

  virtual int translateInput(const std::string &inputName){
    for(unsigned int i = 0; i < inputs.size(); i++){
      if(inputs[i].name == inputName){
        return i;
      }
    }
    if(inputName == "IN_SOURCES"){
      input_sourcesID = addInput(inputName);
      return input_sourcesID;
    }
    else{
      throw new NodeException(this, inputName + " is not supported.", __FILE__, __LINE__);
    }
  }

  void calculate(int output_id, int count, Buffer &out)
  {
    //const Map<int, ObjectRef>& masks = object_cast<Map<int, ObjectRef> >(masks_ptr);
    
    ObjectRef pcms_in = getInput(input_pcmID, count);

    RCPtr<Vector<ObjectRef> > sources_ptr;
    if(input_sourcesID != -1){
      sources_ptr = getInput(input_sourcesID, count);
    }
    else{
      sources_ptr = Vector<ObjectRef>::alloc(0);
    }

    if(typeid(*pcms_in)==typeid(Map<int, ObjectRef>)){

      const Vector<ObjectRef>& sources = *sources_ptr;

      RCPtr<Map<int, ObjectRef> > pcms_ptr = pcms_in;
      const Map<int, ObjectRef>& pcms = *pcms_ptr;

      RCPtr<Source> src;
      int i, j, k;

      for(i=0; i<sources.size(); i++) {
        src = sources[i];
        Map<int, ObjectRef>::const_iterator pcmdata_cit = pcms.find(src->id);
        if(pcmdata_cit != pcms.end()){ // if found Source ID matched PCM data, create new source map element and encode server map element.
          //std::cerr << src->id << " " << pcmdata_cit->first << std::endl;
          //std::cerr << pcmdata_cit->second << std::endl;
          Vector<float>& tmpvec = object_cast<Vector<float> >(pcmdata_cit->second);
          if(sources_map.find(src->id) == sources_map.end()){
            std::vector<float> pcmdata_vec(tmpvec.size());
            for(j=0; j<tmpvec.size(); j++){ pcmdata_vec[j] = tmpvec[j]; }
            std::map<std::string, std::string> source_map;
            std::ostringstream src_ss("");
            //// Source structure(class) variables value append to vorbis comments metadata (MUST need)
            // Basic
            src_ss << src->id;             source_map["id"]            = src_ss.str(); src_ss.str("");
            src_ss << src->x[0];           source_map["x_0"]           = src_ss.str(); src_ss.str("");
            src_ss << src->x[1];           source_map["x_1"]           = src_ss.str(); src_ss.str("");
            src_ss << src->x[2];           source_map["x_2"]           = src_ss.str(); src_ss.str("");
            src_ss << src->power;          source_map["power"]         = src_ss.str(); src_ss.str("");
            // remaining, count and last_found_count is not append to vorbis comments metadata (because, re-framed on receiver node)
            // TF-index
            //src_ss << src->compare_mode;   source_map["compare_mode"]  = src_ss.str(); src_ss.str("");
            if(src->compare_mode == src->DEG){ source_map["compare_mode"] = "DEG"; }
            else{                              source_map["compare_mode"] = "TFINDEX"; }
            src_ss << src->tfindex[0];     source_map["tfindex_0"]     = src_ss.str(); src_ss.str("");
            src_ss << src->tfindex[1];     source_map["tfindex_1"]     = src_ss.str(); src_ss.str("");
            src_ss << src->tfindex[2];     source_map["tfindex_2"]     = src_ss.str(); src_ss.str("");
            src_ss << src->tfindex_max[0]; source_map["tfindex_max_0"] = src_ss.str(); src_ss.str("");
            src_ss << src->tfindex_max[1]; source_map["tfindex_max_1"] = src_ss.str(); src_ss.str("");
            src_ss << src->tfindex_max[2]; source_map["tfindex_max_2"] = src_ss.str(); src_ss.str("");
            src_ss << src->tfindex_min[0]; source_map["tfindex_min_0"] = src_ss.str(); src_ss.str("");
            src_ss << src->tfindex_min[1]; source_map["tfindex_min_1"] = src_ss.str(); src_ss.str("");
            src_ss << src->tfindex_min[2]; source_map["tfindex_min_2"] = src_ss.str(); src_ss.str("");
            //// Julius compatible azimuth/elevation information append to vorbis comments metadata (OPTIONAL)
            //src_ss << (180 / M_PI * atan2f(src->x[1], src->x[0])); source_map["azimuth"]   = src_ss.str(); src_ss.str("");
            //src_ss << (180 / M_PI * asin(src->x[2]));              source_map["elevation"] = src_ss.str(); src_ss.str("");
            //// Julius compatible timestamp information append to vorbis comments metadata (OPTIONAL)
            //timeval tv; gettimeofday(&tv, NULL);
            //src_ss << tv.tv_sec;  source_map["tv_sec"]  = src_ss.str(); src_ss.str("");
            //src_ss << tv.tv_usec; source_map["tv_usec"] = src_ss.str(); src_ss.str("");
            sources_map[src->id] = source_map;

            servers_map[src->id] = new RequestServer::EncodeServer();
            servers_map[src->id]->set_encode_mode(param_i_compression_level, param_i_channel, param_i_sampling_rate, param_i_bits_per_sample, 1024);

            // [NOTE]
            // If you delay update of the metadata (Source information), the set_metadata() method and start() method should not be executed here.
            // Then, it is necessary to separate the write_buffer contained in the reqest_encode. write_buffer method is protected method now.
            servers_map[src->id]->set_metadata(sources_map[src->id]);
            servers_map[src->id]->start();
            servers_map[src->id]->request_encode(pcmdata_vec, pcmdata_vec.size());
          }
          else{
            std::vector<float> pcmdata_vec(param_i_advance);
            for(j=0, k=tmpvec.size()-param_i_advance; k<tmpvec.size(); j++, k++){ pcmdata_vec[j] = tmpvec[k]; }
            servers_map[src->id]->request_encode(pcmdata_vec, pcmdata_vec.size());
          }
          //std::cerr << "[DEBUG] Created new sources_map, servers_map" << std::endl;
        }
        else{
          //std::cerr << "[ERROR] Source is found, but Source ID matched PCM data not found." << std::endl;
        }
      }

      // stop and remove server_map element for terminated Source IDs.
      for(std::map<int, RequestServer::EncodeServer *>::iterator itenc = servers_map.begin(); itenc != servers_map.end();){
        i=sources.size();
        while(i>0){
          src = sources[i-1];
          if(itenc->first==src->id){ break; } i--;
        }
        if(i==0){
          itenc->second->stop();
          std::vector<unsigned char> charbuf = itenc->second->get_encoded_flac();
          //[NOTE] &charbuf[0] was cut after symbol byte 0x00, please use iterator charbuf.begin(), charbuf.end(). Please specify the RANGE explicitly.
          std::string dat; dat.assign(charbuf.begin(), charbuf.end());
          mqttobj->send_message(dat, param_i_qos, /*count=*/0); // count argument was not effective now.
          //[DEBUG code]
          //std::cerr << "Encoded FLAC Data Size: " << charbuf.size() << std::endl;
          //std::cerr << "Encoded FLAC Data:" << std::endl;
          //for(std::vector<unsigned char>::iterator it=charbuf.begin(); it<charbuf.end(); it++){
          //  std::cerr << std::hex << std::setw(2) << ((int) *it & 0xff) << " ";
          //}
          //std::cerr << std::dec << std::endl;
          //[DEBUG code]
          //std::cerr << "Encoded FLAC Data Size: " << dat.size() << std::endl;
          //std::cerr << "Encoded FLAC Data:" << std::endl;
          //for(std::string::iterator it=dat.begin(); it<dat.end(); it++){
          //  std::cerr << std::hex << std::setw(2) << ((int) *it & 0xff) << " ";
          //}
          //std::cerr << std::dec << std::endl;

          if(param_b_enable_save){
            std::ostringstream filename_str;
            std::ofstream ofs;
            filename_str << param_s_filename_prefix << itenc->first << ".flac";
            ofs.open(filename_str.str().c_str(), std::ios::out|std::ios::binary|std::ios::trunc);
            if(ofs){
              ofs.write((char*)&charbuf[0], charbuf.size());
              ofs.close();
            }
          }
          delete itenc->second;
          servers_map.erase(itenc++);
          //std::string mark="end";  mqttobj->send_message(mark, 2, count);
        }
        else{
          itenc++;
          //std::string mark="loop"; mqttobj->send_message(mark, 2, count);
        }
      }
      //while(mqttobj->get_life()>0){ mqttobj->request_loop(); /*std::cerr << ".";*/ } /*std::cerr << " finished." << std::endl; */
      ////for(int loop_i = 5; (loop_i > 0) && (mqttobj->get_life() > prev_life); loop_i--){ std::cerr << "#" << std::endl; mqttobj->request_loop(); std::cerr << "."; } /*std::cerr << " finished." << std::endl; */
      /*if(prev_life > 0){*/ /*prev_life = mqttobj->get_life()*//* - 1*//*;*/ /*std::cerr << "!" << std::endl; }*/
      // remove sources_map element for terminated Source IDs.
      // If you want to change arbitrarily the timing to update the metadata of Source,
      // you MUST want to delete the sources_map after than the servers_map.
      for(std::map<int, std::map<std::string, std::string> >::iterator itsrc = sources_map.begin(); itsrc != sources_map.end();){
        i=sources.size();
        while(i>0){
          src = sources[i-1];
          if(itsrc->first==src->id){ break; } i--;
        }
        if(i==0){
          sources_map.erase(itsrc++);
          //std::string mark="end";  mqttobj->send_message(mark, 2, count);
        }
        else{
          itsrc++;
          //std::string mark="loop"; mqttobj->send_message(mark, 2, count);
        }
      }

      /*
      //[DEBUG code] sources_map print
      bool always_print = false;
      if(always_print==true || sources_map.size()!=0){ std::cerr << "[Debug]: sorces_map" << std::endl; }
      for(std::map<int, std::map<std::string, std::string> >::iterator itsrc = sources_map.begin(); itsrc != sources_map.end(); itsrc++){
        for(std::map<std::string, std::string>::iterator itsrcelem = itsrc->second.begin(); itsrcelem != itsrc->second.end(); itsrcelem++){
          std::cerr << "  " << itsrcelem->first << " " << itsrcelem->second << std::endl;
        }
      }
      */
    }

    else if(typeid(*pcms_in)==typeid(Matrix<float>)){

      const Vector<ObjectRef>& sources = *sources_ptr;

      RCPtr<Matrix<float> > pcms_ptr = pcms_in;

      int i, j, k;
      //std::cerr << "Matrix size: " << pcms_ptr->nrows() << ", " << pcms_ptr->ncols() << std::endl;
      if(pcms_ptr->nrows() >= param_i_channel){
        // first frame create new encode server map element.
        if(count==0){
          //std::vector<float> pcmdata_vec(pcms_ptr->ncols() * pcms_ptr->nrows());
          std::vector<float> pcmdata_vec(pcms_ptr->ncols() * param_i_channel);
          // interleaved
          k=0;
          for(i=0; i<pcms_ptr->ncols(); i++){
            //for(j=0; j<pcms_ptr->nrows(); j++){
            for(j=0; j<param_i_channel; j++){
              pcmdata_vec[k++] = (*pcms_ptr)(j, i);
            }
          }
          servers_map[0] = new RequestServer::EncodeServer();
          servers_map[0]->set_encode_mode(param_i_compression_level, param_i_channel, param_i_sampling_rate, param_i_bits_per_sample, 1024);
          //servers_map[0]->set_metadata("");
          servers_map[0]->start();
          servers_map[0]->request_encode(pcmdata_vec, pcms_ptr->ncols());
        }
        else{
          //std::vector<float> pcmdata_vec(param_i_advance * pcms_ptr->nrows());
          std::vector<float> pcmdata_vec(param_i_advance * param_i_channel);
          // interleaved
          k=0;
          for(i=pcms_ptr->ncols()-param_i_advance; i<pcms_ptr->ncols(); i++){
            //for(j=0; j<pcms_ptr->nrows(); j++){
            for(j=0; j<param_i_channel; j++){
              pcmdata_vec[k++] = (*pcms_ptr)(j, i);
            }
          }
          servers_map[0]->request_encode(pcmdata_vec, param_i_advance);
        }

        // When sending in frame units. (Default on the HARK_FLAC specification)
        std::vector<unsigned char> charbuf = servers_map[0]->get_encoded_flac(65536, true);
        // When sending with fixed byte length buffer. (Additional: processing start may be delayed)
        //std::vector<unsigned char> charbuf = servers_map[0]->get_encoded_flac(512);
        //std::vector<unsigned char> charbuf = servers_map[0]->get_encoded_flac(1024);

        if(charbuf.size() != 0){
        //[NOTE] &charbuf[0] was cut after symbol byte 0x00, please use iterator charbuf.begin(), charbuf.end(). Please specify the RANGE explicitly.
        std::string dat; dat.assign(charbuf.begin(), charbuf.end());
        mqttobj->send_message(dat, param_i_qos, /*count=*/0); // count argument was not effective now.
        //[DEBUG code]
        //std::cerr << "Encoded FLAC Data Size: " << charbuf.size() << std::endl;
        //std::cerr << "Encoded FLAC Data:" << std::endl;
        //for(std::vector<unsigned char>::iterator it=charbuf.begin(); it<charbuf.end(); it++){
        //  std::cerr << std::hex << std::setw(2) << ((int) *it & 0xff) << " ";
        //}
        //std::cerr << std::dec << std::endl;
        //[DEBUG code]
        //std::cerr << "Encoded FLAC Data Size: " << dat.size() << std::endl;
        //std::cerr << "Encoded FLAC Data:" << std::endl;
        //for(std::string::iterator it=dat.begin(); it<dat.end(); it++){
        //  std::cerr << std::hex << std::setw(2) << ((int) *it & 0xff) << " ";
        //}
        //std::cerr << std::dec << std::endl;

        if(param_b_enable_save){
          std::ostringstream filename_str;
          std::ofstream ofs;
          filename_str << param_s_filename_prefix << ".flac";
          ofs.open(filename_str.str().c_str(), std::ios::out|std::ios::binary|std::ios::app);
          if(ofs){
            ofs.write((char*)&charbuf[0], charbuf.size());
            ofs.close();
          }
        }
        } // charbuf.size() != 0
        //std::string mark="end";  mqttobj->send_message(mark, 2, count);
      }
      else{
        // error
      }
      //std::cerr << "[DEBUG] Created new servers_map" << std::endl;
    }

    out[count] = sources_ptr;

  }

  IN_ORDER_NODE_SPEEDUP(PublisherFLAC);

};

//#endif // ENABLE_MQTT

